from __future__ import print_function, division, absolute_import
import os
import shutil

import pandas as pd

from odin.utils import UnitTimer, ctext, ArgController

from sisua.analysis import Posterior, ResultsSheet, get_all_posteriors
from sisua.data.path import EXP_DIR

# ===========================================================================
# Helper
# ===========================================================================
def get_arguments():
  args = ArgController(
  ).add('ds', 'dataset for evaluation'
  ).add('-model', 'name of all model for evaluation, e.g. vae,movae,dca,scvi', 'vae'
  ).add('-ynorm', 'specific normalization mode for protein', ''
  ).add('-yweight', 'specific weight for protein, e.g. 1,10,100', ''
  ).add('-incl', 'list of all keywords for inclusion', ''
  ).add('-excl', 'list of all keywords for exclusion', ''
  ).add('-outpath', 'Output folder to save all the results', ''
  ).parse()

  # ====== checking dataset ====== #
  dataset_name = args.ds

  # ====== checking models ====== #
  models = [i for i in args.models.split(',') if len(i) > 0]
  assert len(models) > 0

  # ====== checking keywords ====== #
  incl = [i for i in args.incl.split(',') if len(i) > 0]
  excl = [i for i in args.incl.split(',') if len(i) > 0]

  # ====== checking protein normalization ====== #
  ynorm = ''
  if args.ynorm == 'prob':
    ynorm = 'Yprob0bernoulli'
  elif args.ynorm == 'raw':
    ynorm = 'Yraw0nb'

  yweight = str(args.yweight)
  yweight = ['%.2g' % float(i)
             for i in yweight.split(',')
             if len(i) > 0]

  # ====== checking output path ====== #
  if args.outpath == '':
    if not os.path.exists('/tmp/results'):
      os.mkdir('/tmp/results')
    outpath = os.path.join('/tmp/results', dataset_name.replace('_', ''))
  else:
    outpath = args.outpath
  outpath = outpath + '_' + '_'.join(models)
  if len(ynorm) > 0:
    outpath += '_' + ynorm
  # incl = ['080spvs1e+02']

  return dataset_name, models, incl, excl, ynorm, yweight, outpath

# ===========================================================================
# Main analyzer
# ===========================================================================
def analyze(dataset_name, models, incl, excl,
            ynorm, yweight, outpath):
  dataset_path = os.path.join(EXP_DIR, dataset_name)
  assert os.path.exists(dataset_path), dataset_path

  if os.path.exists(outpath):
    shutil.rmtree(outpath)
  os.mkdir(outpath)
  score_path = os.path.join(outpath, 'scores.html')

  print("Dataset:", ctext(dataset_name, 'lightyellow'))
  print("Testing models:", ctext(models, 'lightyellow'))
  print("Protein:")
  print(" Yweight:", ctext(yweight, 'yellow'))
  print(" Ynorm:", ctext(ynorm, 'yellow'))
  print("Keywords:")
  print(" Include:", ctext(incl, 'yellow'))
  print(" Exclude:", ctext(excl, 'yellow'))
  print("Saving output at %s" % ctext(outpath, 'lightcyan'))
  print("Saving scores at %s" % ctext(score_path, 'lightcyan'))

  # ====== create the filtering functions ====== #
  def fn_filter(keywords):
    if not any(i in keywords for i in models):
      return False

    if not all(i in keywords for i in incl):
      return False

    if any(i in keywords for i in excl):
      return False

    # only continue with semi-supervised model
    if not ('scvi' in keywords or 'dca' in keywords or 'vae' in keywords):
      if len(ynorm) > 0:
        if keywords[3] != ynorm:
          return False

      if len(yweight) > 0:
        if not any(i == keywords[6].split('spvs')[-1]
                   for i in yweight):
          return False

    return True

  # ====== loading the Posteriors ====== #
  all_posteriors = get_all_posteriors(dataset_path,
      fn_filter=fn_filter,
      show_progress=True
  )

  # Extracting the report
  res = ResultsSheet(all_posteriors, verbose=True)
  print(res)

  # ====== scatter and series ====== #
  res.plot_learning_curves()

  res.plot_imputation_scatter(test=False, pca=True, color_by_library=True)
  res.plot_imputation_scatter(test=True, pca=True, color_by_library=True)

  res.plot_imputation_scatter(test=False, pca=True, color_by_library=False)
  res.plot_imputation_scatter(test=True, pca=True, color_by_library=False)

  res.plot_correlation_marker_pairs(test=False)
  res.plot_correlation_marker_pairs(test=True)

  res.plot_latents_binary_scatter(test=False)
  res.plot_latents_binary_scatter(test=True)

  res.save_plots(outpath, separate_files=True)

  # ====== save scores ====== #
  res.save_scores(score_path)

  # ====== plot scores ====== #
  res.plot_scores('cluster'
  ).plot_scores('pearson'
  ).plot_scores('spearman'
  ).plot_scores('imputation'
  ).plot_scores('classifier'
  )
  res.save_plots(outpath, separate_files=True)

# ===========================================================================
# Main
# ===========================================================================if __name__ == '__main__':
def main():
  args = get_arguments()
  analyze(*args)

if __name__ == '__main__':
  main()



from __future__ import absolute_import, division, print_function

import inspect
import multiprocessing as mpi
import os
import string
from abc import ABCMeta, abstractmethod, abstractproperty
from collections import OrderedDict, defaultdict
from functools import partial
from typing import Iterable, List, Text, Union

import dill
import numpy as np
import tensorflow as tf
from six import add_metaclass, string_types
from tensorflow.python.keras.callbacks import (Callback, CallbackList,
                                               LambdaCallback,
                                               LearningRateScheduler,
                                               ModelCheckpoint)
from tensorflow.python.keras.engine import base_layer_utils
from tensorflow.python.keras.layers import Layer
from tensorflow.python.platform import tf_logging as logging
from tensorflow_probability.python import distributions as tfd
from tqdm import tqdm

from odin.backend.keras_callbacks import EarlyStopping
from odin.bay.distribution_alias import parse_distribution
from odin.bay.distribution_layers import VectorDeterministicLayer
from odin.bay.distributions import concat_distribution
from odin.networks import AdvanceModel
from odin.utils import cache_memory, classproperty
from sisua.data import SingleCellOMIC
from sisua.models import latents as sisua_latents
from sisua.models import networks as sisua_networks

try:
  from hyperopt import hp, fmin, Trials, STATUS_OK, STATUS_FAIL
  from hyperopt.tpe import suggest as tpe_suggest
  from hyperopt.rand import suggest as rand_suggest
  from hyperopt.pyll import scope
except ImportError as e:
  raise RuntimeError(
      "Cannot import hyperopt for hyper-parameters tuning, error: %s" % str(e))


# ===========================================================================
# Helper
# ===========================================================================
def _to_sc_omics(x):
  if isinstance(x, SingleCellOMIC):
    return x
  return SingleCellOMIC(x)


def _to_tfdata(sco: SingleCellOMIC, mask: Union[np.ndarray, None],
               is_semi_supervised, batch_size, shuffle, epochs):
  all_data = []
  for i, j in enumerate(sco):
    if i == 0:  # main data, gene expression
      all_data += [j.X, j.local_mean, j.local_var]
    else:  # semi-supervised data
      all_data.append(j.X)
  # adding mask at the end if semi-supervised
  if is_semi_supervised and mask is not None:
    all_data += [mask]
  # NOTE: from_tensor_slices accept tuple but not list
  ds = tf.data.Dataset.from_tensor_slices(tuple(all_data))
  if shuffle:
    ds = ds.shuffle(1000)
  ds = ds.batch(batch_size).prefetch(tf.data.experimental.AUTOTUNE)
  ds = ds.repeat(epochs)
  # just return a tuple with 1 element to trick keras
  if len(all_data) == 1:
    ds = ds.map(lambda arg: (arg,))
  else:
    ds = ds.map(lambda *args: (args,))
  return ds


def _get_loss_and_distribution(loss):
  # note: loss function always return keepdims=True
  loss = str(loss).lower()
  # ====== simple loss function from tensorflow ====== #
  if loss in dir(tf.losses):
    output_layer = VectorDeterministicLayer
    activation = 'relu'
    loss_fn = lambda y_true, y_pred: tf.expand_dims(tf.losses.get(str(loss))
                                                    (y_true, y_pred),
                                                    axis=-1)
  # ====== probabilistic loss ====== #
  else:
    output_layer = parse_distribution(loss)[0]
    activation = 'linear'
    loss_fn = lambda y_true, y_pred: tf.expand_dims(-y_pred.log_prob(y_true),
                                                    axis=-1)
  # post-processing the loss function to be more universal
  return output_layer, activation, loss_fn


_CACHE_PREDICT = defaultdict(dict)
_MAXIMUM_CACHE_SIZE = 2


# ===========================================================================
# SingleCell model
# ===========================================================================
@add_metaclass(ABCMeta)
class SingleCellModel(AdvanceModel):
  """

  I/O
  ---
  >>> import dill
  >>> c, w = model.get_config(), model.get_weights()
  >>> with open('/tmp/model', 'wb') as f:
  >>>   dill.dump((c, w), f)
  ...
  >>> with open('/tmp/model', 'rb') as f:
  >>>   (c, w) = dill.load(f)
  >>> model = SingleCellModel.from_config(c)
  >>> model.set_weights(w)


  Note
  ----
  It is recommend to call `tensorflow.random.set_seed` for reproducible results
  """

  def __init__(self,
               units,
               xdist,
               dispersion,
               parameters,
               kl_analytic=True,
               kl_weight=1.,
               kl_warmup=50,
               log_norm=True,
               seed=8,
               name=None):
    if name is None:
      name = self.__class__.__name__
    parameters.update(locals())
    super(SingleCellModel, self).__init__(parameters=parameters, name=name)
    self._epochs = self.add_weight(
        name='epochs',
        shape=(),
        initializer=tf.initializers.Constant(value=0.),
        trainable=False,
        dtype='float32')
    # parameters for ELBO
    self._kl_analytic = bool(kl_analytic)
    self._kl_weight = tf.convert_to_tensor(kl_weight, dtype='float32')
    self._kl_warmup = tf.convert_to_tensor(kl_warmup, dtype='float32')
    self._seed = int(seed)
    self._is_fitting = False
    self._corruption_rate = 0
    self._corruption_dist = 'binomial'
    self._log_norm = bool(log_norm)
    # dispersion
    dispersion = str(dispersion).lower()
    assert dispersion in ('full', 'share', 'single')
    self.dispersion = dispersion
    # parsing the units
    if not isinstance(units, (tuple, list)):
      units = [units]
    units = [int(i) for i in units]
    self._units = units
    # parsing the input distribution
    if not isinstance(xdist, (tuple, list)):
      xdist = [xdist]
    # check length of units is matching xdist,
    # i.e. consistent number of outputs
    assert len(units) == len(xdist), \
      "Given %d output(s) in 'units' which differnt from  " % len(units) + \
        "%d distribution(s) given in 'xdist'" % len(xdist)
    # check if mRNA output is zero inflated
    if 'zi' == xdist[0][:2].lower():
      self._is_zero_inflated = True
    else:
      self._is_zero_inflated = False
    # check if is semi-supervised
    self._is_semi_supervised = True if len(xdist) > 1 else False
    # for initializing the output layers
    self.xdist = []
    self.xloss = []
    self.xactiv = []
    for i in xdist:
      fn_dist, activation, fn_loss = _get_loss_and_distribution(i)
      self.xloss.append(fn_loss)
      self.xdist.append(fn_dist)
      self.xactiv.append(activation)

  @property
  def units(self):
    """ Always return a list of units for all outputs """
    return tuple(self._units)

  @property
  def n_outputs(self):
    return len(self.xdist)

  @property
  def log_norm(self):
    return self._log_norm

  @property
  def custom_objects(self):
    return [sisua_latents, sisua_networks]

  @property
  def is_zero_inflated(self):
    return self._is_zero_inflated

  @property
  def is_semi_supervised(self):
    return self._is_semi_supervised

  @property
  def corruption_rate(self):
    return self._corruption_rate

  @property
  def corruption_dist(self):
    return self._corruption_dist

  @property
  def epochs(self):
    return int(self._epochs.numpy())

  @property
  def kl_analytic(self):
    return self._kl_analytic

  @property
  def kl_warmup(self):
    return self._kl_warmup

  @property
  def kl_weight(self):
    warmup_weight = tf.minimum(
        tf.maximum(self._epochs, 1.) / self.kl_warmup, 1.)
    return warmup_weight * self._kl_weight

  @property
  def seed(self):
    return self._seed

  @property
  def parameters(self):
    return self._parameters

  @abstractmethod
  def _call(self, x, lmean, lvar, t, y, masks, training=None, n_samples=None):
    """
    x : [batch_size, n_genes]
      input gene-expression matrix (multiple inputs can be given)
    lmean : [batch_size]
      mean of library size
    lvar : [batch_size]
      variance of library size
    t : [batch_size, n_genes]
      target for reconstruction of gene-expression matrix (can be different
      from `x`)
    y : [batch_size, n_protein]
      input for semi-supervised learning
    masks : [batch_size, 1]
      binary mask for semi-supervised training
    training : `bool`
      flag mark training progress
    n_samples : {`None`, `int`}
      number of MCMC samples
    """
    raise NotImplementedError

  def call(self, inputs, training=None, n_samples=None):
    # check arguments
    if n_samples is None:
      n_samples = 1
    if not isinstance(inputs, (tuple, list)):
      inputs = [inputs]
    else:
      inputs = list(inputs)

    # eager mode, the Model is called directly
    # lmean and lvar is not given
    if (len(inputs) >= 3 and
        (inputs[1].shape[1] != 1 or inputs[2].shape[1] != 1)) or \
      len(inputs) < 3:
      if isinstance(inputs[0], SingleCellOMIC):
        inputs = [inputs[0].X, inputs[0].local_mean, inputs[0].local_var
                 ] + inputs[1:]
      else:
        inputs = [
            inputs,
            tf.zeros(shape=(inputs.shape[0], 1)),
            tf.zeros(shape=(inputs.shape[0], 1))
        ] + inputs[1:]

    # Preprocessing the inputs for semi-supervised training
    x, lmean, lvar = inputs[:3]
    if self._is_fitting:
      y = list(inputs[3:-1])  # y1, y2, ...
      masks = [inputs[-1]] * len(y)  # same mask for all y
    else:
      y = list(inputs[3:])  # y1, y2, ...
      masks = []

    # check log normalization
    t = x
    if self.log_norm:
      x = tf.math.log1p(x)

    # postprocessing the returns
    outputs, latents = self._call(x, lmean, lvar, t, y, masks, training,
                                  n_samples)
    if not self.is_semi_supervised and isinstance(outputs, (tuple, list)):
      outputs = outputs[0]
    return outputs, latents

  def evaluate(self, inputs, n_samples=1, batch_size=128, verbose=1):
    raise Exception(
        "This method is not support, please use sisua.analysis.Posterior")

  def predict(self,
              inputs,
              n_samples=1,
              batch_size=64,
              apply_corruption=False,
              enable_cache=True,
              verbose=1):
    """
    Parameters
    ----------
    apply_corruption : `bool` (default=`False`)
      if `True` applying corruption on data before prediction to match the
      condition during fitting.
    enable_cache : `bool` (default=`True`)
      if `True` store the "footprint" of the input arguments to return the
      cached outputs

    Return
    ------
    X : `Distribution` or tuple of `Distribution`
      output distribution, multiple distribution is return in case of
      multiple outputs
    Z : `Distribution` or tuple of `Distribution`
      latent distribution, multiple distribution is return in case of
      multiple latents
    """
    if not isinstance(inputs, (tuple, list)):
      inputs = [inputs]
    inputs = [_to_sc_omics(i) for i in inputs]
    assert len(inputs) == 1, \
      "During prediction phase, only the mRNA gene expression is provided, " +\
        "this is strict regulation for all models!"

    # checking the cache, this mechanism will significantly improve speed
    # during monitoring of fitting process
    self_id = id(self)
    footprint = ''.join([str(id(i.X)) for i in inputs]) + \
      str(n_samples) + str(apply_corruption) + str(self.epochs)
    if enable_cache and footprint in _CACHE_PREDICT[id(self)]:
      return _CACHE_PREDICT[self_id][footprint]

    # applying corruption for testing
    if apply_corruption and self.corruption_rate is not None:
      inputs = [
          data.corrupt(corruption_rate=self.corruption_rate,
                       corruption_dist=self.corruption_dist,
                       inplace=False) if idx == 0 else data
          for idx, data in enumerate(inputs)
      ]
    n = inputs[0].shape[0]
    data = _to_tfdata(inputs,
                      None,
                      self.is_semi_supervised,
                      batch_size,
                      shuffle=False,
                      epochs=1)

    kw = {'n_samples': int(n_samples)}
    if 'n_samples' not in self._call_fn_args:
      del kw['n_samples']

    X, Z = [], []
    for inputs in tqdm(data,
                       desc="Predicting",
                       total=int(np.ceil(n / batch_size)),
                       disable=not bool(verbose)):
      # the _to_tfddata will return (x, y) tuple for `fit` methods,
      # y=None and we only need x here.
      x, z = self(inputs[0], training=False, **kw)
      X.append(x)
      Z.append(z)

    if isinstance(x, (tuple, list)):
      merging_axis = 0 if x[0].batch_shape.ndims == 1 else 1
    else:
      merging_axis = 0 if x.batch_shape.ndims == 1 else 1
    # multiple outputs
    if isinstance(X[0], (tuple, list)):
      X = tuple([
          concat_distribution([x[idx]
                               for x in X], axis=merging_axis)
          for idx in range(len(X[0]))
      ])
    # single output
    else:
      X = concat_distribution(X, axis=merging_axis)

    # multiple latents
    if isinstance(Z[0], (tuple, list)):
      Z = tuple([
          concat_distribution([z[idx]
                               for z in Z], axis=0)
          for idx in range(len(Z[0]))
      ])
    else:
      Z = concat_distribution(Z, axis=0)
    # cache and return
    if enable_cache:
      _CACHE_PREDICT[self_id][footprint] = (X, Z)
      # LIFO
      if len(_CACHE_PREDICT[self_id]) > _MAXIMUM_CACHE_SIZE:
        key = list(_CACHE_PREDICT[self_id].keys())[0]
        del _CACHE_PREDICT[self_id][key]
    return X, Z

  def fit(
      self,
      inputs: Union[SingleCellOMIC, Iterable[SingleCellOMIC]],
      optimizer: Union[Text, tf.optimizers.Optimizer] = 'adam',
      learning_rate=1e-4,
      clipnorm=100,
      n_samples=1,
      semi_percent=0.8,
      semi_weight=10,
      corruption_rate=0.25,
      corruption_dist='binomial',
      batch_size=64,
      epochs=2,
      callbacks=None,
      validation_split=0.1,
      validation_freq=1,
      min_delta=0.5,  # for early stopping
      patience=25,  # for early stopping
      allow_rollback=True,  # for early stopping
      terminate_on_nan=True,  # for early stopping
      checkpoint=None,
      shuffle=True,
      verbose=1):
    """ This fit function is the combination of both
    `Model.compile` and `Model.fit` """
    # check signature of `call` function
    # inputs, training=None, n_samples=1
    specs = inspect.getfullargspec(self.call)
    if specs.args != ['self', 'inputs', 'training', 'n_samples']:
      raise ValueError(
          "call method must have following arguments %s; bug given %s" %
          (['self', 'inputs', 'training', 'n_samples'], specs.args))

    # check arguments
    assert 0.0 <= semi_percent <= 1.0
    if validation_split <= 0:
      raise ValueError("validation_split must > 0")

    # store the corruption rate for later use
    self._corruption_dist = corruption_dist
    self._corruption_rate = corruption_rate

    # prepare input data
    if not isinstance(inputs, (tuple, list)):
      inputs = [inputs]
    inputs = [_to_sc_omics(i) for i in inputs]

    # corrupting the data (only the main data, the semi supervised data
    # remain clean)
    if corruption_rate > 0:
      inputs = [
          data.corrupt(corruption_rate=corruption_rate,
                       corruption_dist=corruption_dist,
                       inplace=False) if idx == 0 else data
          for idx, data in enumerate(inputs)
      ]
    train, valid = [], []
    for i in inputs:
      tr, va = i.split(seed=self.seed, train_percent=1 - validation_split)
      train.append(tr)
      valid.append(va)

    # generate training mask for semi-supervised learning
    rand = np.random.RandomState(seed=self.seed)
    n = train[0].shape[0]
    train_mask = np.zeros(shape=(n, 1), dtype='float32')
    train_mask[rand.permutation(n)[:int(semi_percent * n)]] = 1
    train_mask = train_mask * semi_weight
    valid_mask = np.ones(shape=(valid[0].shape[0], 1), dtype='float32')

    # calculate the steps
    assert len(set(i.shape[0] for i in train)) == 1
    assert len(set(i.shape[0] for i in valid)) == 1
    steps_per_epoch = int(np.ceil(train[0].shape[0] / batch_size))
    validation_steps = int(np.ceil(valid[0].shape[0] / batch_size))

    # convert to tensorflow Dataset
    train_data = _to_tfdata(train, train_mask, self.is_semi_supervised,
                            batch_size, shuffle, epochs)
    valid_data = _to_tfdata(valid, valid_mask, self.is_semi_supervised,
                            batch_size, shuffle, epochs)

    # prepare callback
    update_epoch = LambdaCallback(
        on_epoch_end=lambda *args, **kwargs: self._epochs.assign_add(1))
    if callbacks is None:
      callbacks = [update_epoch]
    elif isinstance(callbacks, Callback):
      callbacks = [callbacks, update_epoch]
    else:
      callbacks = list(callbacks)
      callbacks.append(update_epoch)
    # add early stopping
    if patience >= 0:
      callbacks.append(
          EarlyStopping(monitor='val_loss',
                        min_delta=min_delta,
                        patience=patience,
                        verbose=verbose,
                        mode='min',
                        baseline=None,
                        terminate_on_nan=bool(terminate_on_nan),
                        restore_best_weights=bool(allow_rollback)))
    # checkpoint
    if checkpoint is not None:
      callbacks.append(
          ModelCheckpoint(filepath=str(checkpoint),
                          monitor='val_loss',
                          verbose=verbose,
                          save_weights_only=True,
                          save_freq='epoch',
                          save_best_only=True,
                          load_weights_on_restart=False,
                          mode='min'))
    # automatically set inputs as validation set for missing inputs
    # metrical callbacks
    from sisua.analysis.sc_metrics import SingleCellMetric
    from sisua.analysis.sc_monitor import SingleCellMonitor
    for cb in callbacks:
      if isinstance(cb, (SingleCellMetric, SingleCellMonitor)) and \
        cb.inputs is None:
        # only mRNA expression is needed during evaluation (prediction)
        cb.inputs = valid[0]
    # reorganize so the SingleCellMonitor is at the end, hence, all the
    # metrics are computed before the monitor start plotting
    cb_others = [
        cb for cb in callbacks if not isinstance(cb, SingleCellMonitor)
    ]
    cb_monitor = [cb for cb in callbacks if isinstance(cb, SingleCellMonitor)]
    callbacks = cb_others + cb_monitor

    # start training loop
    org_fn = self.call
    if 'n_samples' in self._call_fn_args:
      self.call = partial(self.call, n_samples=n_samples)
    self._is_fitting = True

    # prepare the logging
    curr_log = logging.get_verbosity()
    logging.set_verbosity(logging.ERROR)
    # compiling and setting the optimizer
    if not self._is_compiled and self.optimizer is None:
      # prepare optimizer
      if isinstance(optimizer, string_types):
        config = dict(learning_rate=learning_rate)
        if clipnorm is not None:
          config['clipnorm'] = clipnorm
        optimizer = tf.optimizers.get({
            'class_name': optimizer,
            'config': config
        })
      elif isinstance(optimizer, tf.optimizers.Optimizer):
        pass
      elif isinstance(optimizer, type) and issubclass(optimizer,
                                                      tf.optimizers.Optimizer):
        optimizer = optimizer(learning_rate=learning_rate) \
          if clipnorm is None else \
          optimizer(learning_rate=learning_rate, clipnorm=clipnorm)
      else:
        raise ValueError("No support for optimizer: %s" % str(optimizer))
      super(SingleCellModel, self).compile(optimizer,
                                           experimental_run_tf_function=False)
    # fitting
    super(SingleCellModel, self).fit(x=train_data,
                                     validation_data=valid_data,
                                     validation_freq=validation_freq,
                                     callbacks=callbacks,
                                     initial_epoch=self.epochs,
                                     steps_per_epoch=steps_per_epoch,
                                     validation_steps=validation_steps,
                                     epochs=self.epochs + epochs,
                                     verbose=verbose)
    logging.set_verbosity(curr_log)

    # reset to original state
    self.call = org_fn
    self._is_fitting = False

  def __repr__(self):
    return self.__str__()

  def __str__(self):
    return "<[%s]%s fitted:%s epoch:%s semi:%s>" % (
        self.__class__.__name__, self.name, self.epochs > 0, self.epochs,
        self.is_semi_supervised)

  @classproperty
  def id(cls):
    class_name = cls.__name__
    name = ''
    for i in class_name:
      if i.isupper():
        name += i
    return name.lower()

  @classmethod
  def fit_hyper(
      cls,
      inputs: Union[SingleCellOMIC, Iterable[SingleCellOMIC]],
      params: dict = {
          'nlayers': scope.int(hp.choice('nlayers', [1, 2, 3, 4])),
          'hdim': scope.int(hp.choice('hdim', [32, 64, 128, 256, 512])),
          'zdim': scope.int(hp.choice('zdim', [32, 64, 128, 256, 512])),
      },
      loss_name: Text = 'val_loss',
      max_evals: int = 100,
      model_kwargs: dict = {},
      fit_kwargs: dict = {
          'epochs': 64,
          'batch_size': 128
      },
      algorithm: Text = 'bayes',
      seed: int = 8,
      save_path: Text = '/tmp/{model:s}_{data:s}_{loss:s}_{params:s}.hp',
      override: bool = True,
      verbose: bool = False):
    """ Hyper-parameters optimization for given SingleCellModel
    Parameters
    ----------
    model_kwargs : `dict`
      keyword arguments for model construction
    fit_kwargs : `dict`
      keyword arguments for `fit` method

    Example
    -------
    >>> callbacks = [
    >>>     NegativeLogLikelihood(),
    >>>     ImputationError(),
    >>>     CorrelationScores(extras=y_train)
    >>> ]
    >>> i, j = DeepCountAutoencoder.fit_hyper(x_train,
    >>>                                       kwargs={'loss': 'zinb'},
    >>>                                       fit_kwargs={
    >>>                                           'callbacks': callbacks,
    >>>                                           'epochs': 64,
    >>>                                           'batch_size': 128
    >>>                                       },
    >>>                                       loss_name='pearson_mean',
    >>>                                       algorithm='bayes',
    >>>                                       max_evals=100,
    >>>                                       seed=8)
    """
    if isinstance(loss_name, string_types):
      loss_name = [loss_name]
    loss_name = [str(i) for i in loss_name]

    algorithm = str(algorithm.lower())
    assert algorithm in ('rand', 'grid', 'bayes'), \
      "Only support 3 algorithm: rand, grid and bayes; given %s" % algorithm
    # force to turn of keras verbose, it is a big mess to show
    # 2 progress bar at once
    fit_kwargs.update({'verbose': 0})

    # remove unncessary params
    args = inspect.getfullargspec(cls.__init__)
    params = {i: j for i, j in params.items() if i in args.args}

    # processing save_path
    fmt = {}
    for (_, key, spec, _) in string.Formatter().parse(save_path):
      if spec is not None:
        fmt[key] = None
    if isinstance(inputs, (tuple, list)):
      dsname = inputs[0].name if hasattr(inputs[0],
                                         'name') else 'x%d' % len(inputs)
    else:
      dsname = inputs.name if hasattr(inputs, 'name') else 'x'
    kw = {
        'model':
            cls.id,
        'data':
            dsname.replace('_', ''),
        'loss':
            '_'.join(
                [i.replace('val_', '').replace('_', '') for i in loss_name]),
        'params':
            '_'.join(sorted([i.replace('_', '') for i in params.keys()]))
    }
    kw = {i: j for i, j in kw.items() if i in fmt}
    save_path = save_path.format(**kw)
    if os.path.exists(save_path) and not override:
      raise RuntimeError("Cannot override path: %s" % save_path)

    # ====== verbose mode ====== #
    if verbose:
      print(" ======== Tunning: %s ======== " % cls.__name__)
      print("Save path:", save_path)
      print("Model config:", model_kwargs)
      print("Fit config  :", fit_kwargs)
      print("Loss name   :", loss_name)
      print("Algorithm   :", algorithm)
      print("Max evals   :", max_evals)
      print("Search space:")
      for i, j in params.items():
        print("  ", i, j)

    def fit_and_evaluate(*args):
      kw = args[0]
      kw.update(model_kwargs)
      obj = cls(**kw)
      obj.fit(inputs, **fit_kwargs)
      history = obj.history.history
      all_loss = [history[name] for name in loss_name]
      # get min, variance and status, if NaN set to Inf
      loss = 0
      loss_variance = 0
      is_nan = False
      for l in all_loss:
        if np.any(np.isnan(l)):
          is_nan = True
          loss = np.inf
          loss_variance = np.inf
          break
        else:  # first epoch doesn't count
          loss += np.min(l[1:])
          loss_variance += np.var(l[1:])
      loss = loss / len(all_loss)
      loss_variance = loss_variance / len(all_loss)
      return {
          'loss': loss,
          'loss_variance': loss_variance,
          'history': history,
          'status': STATUS_FAIL if is_nan else STATUS_OK,
      }

    def hyperopt_run():
      trials = Trials()
      results = fmin(fit_and_evaluate,
                     space=params,
                     algo=tpe_suggest if algorithm == 'bayes' else rand_suggest,
                     max_evals=int(max_evals),
                     trials=trials,
                     rstate=np.random.RandomState(seed),
                     verbose=verbose)
      history = []
      for t in trials:
        r = t['result']
        history.append({
            'loss': r['loss'],
            'loss_variance': r['loss_variance'],
            'params': {i: j[0] for i, j in t['misc']['vals'].items()},
            'history': r['history'],
            'status': r['status'],
        })
      with open(save_path, 'wb') as f:
        print("Saving hyperopt results to: %s" % save_path)
        dill.dump((results, history), f)

    p = mpi.Process(target=hyperopt_run)
    p.start()
    p.join()

    try:
      with open(save_path, 'rb') as f:
        results, history = dill.load(f)
    except FileNotFoundError:
      results, history = {}, {}

    if verbose:
      print("Best:", results)
    return results, history



from typing import Dict, Optional, Text

from sisua.models import latents, networks
from sisua.models.autoencoder import *
from sisua.models.base import *
from sisua.models.scscope_models import *
from sisua.models.scvi_models import *
from sisua.models.semi_supervised import *
from sisua.models.variational_autoencoder import *


def get_model(identifier: Optional[Text] = None) -> Dict:
  from sisua import models
  from types import ModuleType
  from typing import Type
  all_models = {}
  for name in dir(models):
    mod = getattr(models, name)
    if isinstance(mod, ModuleType):
      if 'sisua/models' in mod.__file__:
        for subname in dir(mod):
          clz = getattr(mod, subname)
          if isinstance(clz, Type) and issubclass(
              clz, SingleCellModel) and clz != SingleCellModel:
            all_models[clz.id] = clz
  if identifier is not None:
    return all_models[identifier]
  return all_models



from __future__ import absolute_import, division, print_function

import base64
import os
import pickle
import shutil

import numpy as np

from odin.fuel import Dataset
from odin.utils import get_file
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR
from sisua.data.utils import remove_allzeros_columns, save_to_dataset

_URL_LYMPHOID = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL3BibWNlY2NfbHkubnB6\n'
_URL_MYELOID = None
_URL_FULL = None


def read_PBMCeec(subset, override=False, verbose=False, filtered_genes=False):
  subset = str(subset).strip().lower()
  if subset not in ('ly', 'my', 'full'):
    raise ValueError(
        "subset can only be 'ly'-lymphoid and 'my'-myeloid or 'full'")
  if subset in ('my', 'full'):
    raise NotImplementedError("No support for subset: %s - PBMCecc" % subset)

  download_path = os.path.join(DOWNLOAD_DIR, "PBMCecc_%s_original" % subset)
  if not os.path.exists(download_path):
    os.mkdir(download_path)

  preprocessed_path = os.path.join(
      PREPROCESSED_BASE_DIR,
      'PBMCecc_%s_preprocessed' % (subset + ('' if filtered_genes else 'full')))

  if override and os.path.exists(preprocessed_path):
    shutil.rmtree(preprocessed_path)
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)

  # ******************** preprocessed ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    # ====== full ====== #
    if subset == 'full':
      pass
    # ====== ly and my ====== #
    else:
      url = str(
          base64.decodebytes(_URL_LYMPHOID if subset == 'ly' else _URL_MYELOID),
          'utf-8')
      base_name = os.path.basename(url)
      get_file(fname=base_name,
               origin=url,
               outdir=download_path,
               verbose=verbose)
      # ====== extract the data ====== #
      data = np.load(os.path.join(download_path, base_name))
      X_row = data['X_row']
      y = data['y']
      y_col = data['y_col']
      if filtered_genes:
        X = data['X_var']
        X_col = data['X_var_col']
      else:
        X = data['X_full']
        X_col = data['X_full_col']
      cell_types = None

    # ====== save everything ====== #
    X, X_col = remove_allzeros_columns(matrix=X,
                                       colname=X_col,
                                       print_log=verbose)
    assert X.shape == (len(X_row), len(X_col))
    assert len(X) == len(y)
    assert y.shape[1] == len(y_col)

    if cell_types is not None:
      with open(os.path.join(preprocessed_path, 'cell_types'), 'wb') as f:
        pickle.dump(cell_types, f)

    save_to_dataset(preprocessed_path,
                    X,
                    X_col,
                    y,
                    y_col,
                    rowname=X_row,
                    print_log=verbose)

  # ******************** read preprocessed data ******************** #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds



from __future__ import absolute_import, division, print_function

import copy
import os
from collections import OrderedDict
from numbers import Number
from typing import Tuple

import numpy as np

from odin.stats import describe, sparsity_percentage, train_valid_test_split
from odin.utils import cache_memory, ctext, one_hot
from sisua.data import normalization_recipes
from sisua.data.const import UNIVERSAL_RANDOM_SEED
from sisua.data.single_cell_dataset import (SingleCellOMIC,
                                            apply_artificial_corruption,
                                            get_library_size)
from sisua.data.utils import (get_gene_id2name, standardize_protein_name,
                              validating_dataset)


def get_dataset_meta():
  """
  Return
  ------
  dictionary : dataset name -> loading_function()
  """
  from sisua.data.data_loader.dataset10x import read_dataset10x_cellexp
  from sisua.data.data_loader.pbmc_CITEseq import read_CITEseq_PBMC
  from sisua.data.data_loader.cbmc_CITEseq import read_CITEseq_CBMC
  from sisua.data.data_loader.mnist import read_MNIST
  from sisua.data.data_loader.fashion_mnist import read_fashion_MNIST
  from sisua.data.data_loader.facs_gene_protein import read_FACS, read_full_FACS
  from sisua.data.data_loader.scvi_datasets import (read_Cortex, read_Hemato,
                                                    read_PBMC, read_Retina)
  from sisua.data.data_loader.pbmc8k import read_PBMC8k
  from sisua.data.data_loader.pbmcecc import read_PBMCeec
  from sisua.data.experimental_data.pbmc_8k_ecc_ly import (
      read_PBMCcross_ecc_8k, read_PBMCcross_remove_protein)
  from sisua.data.data_loader.centenarian import read_centenarian
  data_meta = {
      "centenarian":
          read_centenarian,
      # ====== PBMC 10x ====== #
      "neuron10k":
          lambda override=False, verbose=False: read_dataset10x_cellexp(
              name='neuron_10k_v3',
              spec='filtered',
              override=override,
              verbose=verbose),
      "heart10kv3":
          lambda override=False, verbose=False: read_dataset10x_cellexp(
              name='heart_10k_v3',
              spec='filtered',
              override=override,
              verbose=verbose),
      "cellvdj":
          lambda override=False, verbose=False: read_dataset10x_cellexp(
              name='cellvdj',
              spec='filtered',
              override=override,
              verbose=verbose),
      'memoryt':
          lambda override=False, verbose=False: read_dataset10x_cellexp(
              name='memory_t',
              spec='filtered',
              override=override,
              verbose=verbose),
      'pbmc5k':
          lambda override=False, verbose=False: read_dataset10x_cellexp(
              name='5k_pbmc_protein_v3',
              spec='filtered',
              override=override,
              verbose=verbose),
      'pbmc5kgem':
          lambda override=False, verbose=False: read_dataset10x_cellexp(
              name='5k_pbmc_protein_v3_nextgem',
              spec='filtered',
              override=override,
              verbose=verbose),

      # ====== pbmc 8k ====== #
      'pbmc8klyfull':
          lambda override=False, verbose=False: read_PBMC8k(subset='ly',
                                                            override=override,
                                                            verbose=verbose,
                                                            filtered_genes=False
                                                           ),
      'pbmc8kmyfull':
          lambda override=False, verbose=False: read_PBMC8k(subset='my',
                                                            override=override,
                                                            verbose=verbose,
                                                            filtered_genes=False
                                                           ),
      'pbmc8kly':
          lambda override=False, verbose=False: read_PBMC8k(subset='ly',
                                                            override=override,
                                                            verbose=verbose,
                                                            filtered_genes=True
                                                           ),
      'pbmc8kmy':
          lambda override=False, verbose=False: read_PBMC8k(subset='my',
                                                            override=override,
                                                            verbose=verbose,
                                                            filtered_genes=True
                                                           ),
      'pbmc8k':
          lambda override=False, verbose=False: read_PBMC8k(subset='full',
                                                            override=override,
                                                            verbose=verbose,
                                                            filtered_genes=True
                                                           ),
      'pbmc8kfull':
          lambda override=False, verbose=False: read_PBMC8k(subset='full',
                                                            override=override,
                                                            verbose=verbose,
                                                            filtered_genes=False
                                                           ),

      # ====== PBMC ECC ====== #
      'pbmcecclyfull':
          lambda override=False, verbose=False: read_PBMCeec(subset='ly',
                                                             override=override,
                                                             verbose=verbose,
                                                             filtered_genes=
                                                             False),
      # 'pbmcecc_myfull': lambda override: read_PBMCeec(subset='my', override=override, filtered_genes=False),
      'pbmceccly':
          lambda override=False, verbose=False: read_PBMCeec(subset='ly',
                                                             override=override,
                                                             verbose=verbose,
                                                             filtered_genes=True
                                                            ),
      # 'pbmcecc_my': lambda override: read_PBMCeec(subset='my', override=override, filtered_genes=True),
      # 'pbmcecc': lambda override: read_PBMCeec(subset='full', override=override, filtered_genes=True),
      # 'pbmcecc_full': lambda override: read_PBMCeec(subset='full', override=override, filtered_genes=False),

      # ====== cross PBMC ====== #
      'cross8klyfull':
          lambda override=False, verbose=False: read_PBMCcross_ecc_8k(
              subset='ly',
              return_ecc=False,
              override=override,
              verbose=verbose,
              filtered_genes=False),
      'cross8kly':
          lambda override=False, verbose=False: read_PBMCcross_ecc_8k(
              subset='ly',
              return_ecc=False,
              override=override,
              verbose=verbose,
              filtered_genes=True),
      'crossecclyfull':
          lambda override=False, verbose=False: read_PBMCcross_ecc_8k(
              subset='ly',
              return_ecc=True,
              override=override,
              verbose=verbose,
              filtered_genes=False),
      'crosseccly':
          lambda override=False, verbose=False: read_PBMCcross_ecc_8k(
              subset='ly',
              return_ecc=True,
              override=override,
              verbose=verbose,
              filtered_genes=True),
      'cross8knocd4':
          lambda override=False, verbose=False: read_PBMCcross_remove_protein(
              subset='ly',
              return_ecc=False,
              override=override,
              verbose=verbose,
              filtered_genes=True,
              remove_protein='CD4'),
      'crosseccnocd4':
          lambda override=False, verbose=False: read_PBMCcross_remove_protein(
              subset='ly',
              return_ecc=True,
              override=override,
              verbose=verbose,
              filtered_genes=True,
              remove_protein='CD4'),
      'cross8knocd8':
          lambda override=False, verbose=False: read_PBMCcross_remove_protein(
              subset='ly',
              return_ecc=False,
              override=override,
              verbose=verbose,
              filtered_genes=True,
              remove_protein='CD8'),
      'crosseccnocd8':
          lambda override=False, verbose=False: read_PBMCcross_remove_protein(
              subset='ly',
              return_ecc=True,
              override=override,
              verbose=verbose,
              filtered_genes=True,
              remove_protein='CD8'),
      'cross8knocd48':
          lambda override=False, verbose=False: read_PBMCcross_remove_protein(
              subset='ly',
              return_ecc=False,
              override=override,
              verbose=verbose,
              filtered_genes=True,
              remove_protein=['CD4', 'CD8']),
      'crosseccnocd48':
          lambda override=False, verbose=False: read_PBMCcross_remove_protein(
              subset='ly',
              return_ecc=True,
              override=override,
              filtered_genes=True,
              remove_protein=['CD4', 'CD8']),
      'cross8konlycd8':
          lambda override=False, verbose=False: read_PBMCcross_remove_protein(
              subset='ly',
              return_ecc=False,
              override=override,
              verbose=verbose,
              filtered_genes=True,
              remove_protein=['CD3', 'CD4', 'CD16', 'CD56', 'CD19']),
      # ====== CITEseq ====== #
      'pbmcciteseq':
          read_CITEseq_PBMC,
      'cbmcciteseq':
          read_CITEseq_CBMC,
      'pbmc5000':
          lambda override=False, verbose=False: read_CITEseq_PBMC(
              override=override, verbose=verbose, version_5000genes=True),

      # ====== MNIST ====== #
      'mnist':
          read_MNIST,
      'fmnist':
          read_fashion_MNIST,

      # ====== FACS ====== #
      'facs7':
          lambda override=False, verbose=False: read_full_FACS(
              override=override, verbose=verbose),
      'facs5':
          lambda override=False, verbose=False: read_FACS(
              n_protein=5, override=override, verbose=verbose),
      'facs2':
          lambda override=False, verbose=False: read_FACS(
              n_protein=2, override=override, verbose=verbose),

      # ====== other fun ====== #
      'pbmcscvi':
          read_PBMC,
      'cortex':
          read_Cortex,
      'retina':
          read_Retina,
      'hemato':
          read_Hemato,
  }
  import re
  pattern = re.compile('\w*')
  for name in data_meta.keys():
    assert pattern.match(name) and '_' not in name
  return data_meta


def get_dataset_summary(return_html=False):
  from sisua.data.utils import standardize_protein_name
  all_datasets = []
  for name, fn in sorted(get_dataset_meta().items()):
    ds = fn(override=False)
    info = OrderedDict([
        ('Keyword', name),
        ('#Cells', ds['X'].shape[0]),
        ('#Genes', ds['X'].shape[1]),
        ('#Labels', ds['y'].shape[1]),
        ('Binary', sorted(np.unique(ds['y'].astype('float32'))) == [0., 1.]),
        ('Labels', ', '.join([standardize_protein_name(i) for i in ds['y_col']
                             ])),
    ])
    all_datasets.append(info)
  df = pd.DataFrame(all_datasets)
  if return_html:
    return df.to_html()
  return df


def get_dataset(dataset_name, override=False,
                verbose=False) -> Tuple[SingleCellOMIC, SingleCellOMIC]:
  """ Check `get_dataset_meta` for more information

  Return
  ------
  dataset: `odin.fuel.dataset.Dataset` contains original data
  """
  data_meta = get_dataset_meta()
  # ====== special case: get all dataset ====== #
  dataset_name = str(dataset_name).lower().strip()
  if dataset_name not in data_meta:
    raise RuntimeError(
        'Cannot find dataset with name: "%s", all dataset include: %s' %
        (dataset_name, ",".join(list(data_meta.keys()))))
  ds = data_meta[dataset_name](override=override, verbose=verbose)
  validating_dataset(ds)
  # ******************** return ******************** #
  var = {'geneid': ds['X_col']}
  if 'X_col_name' in ds:
    var['genename'] = ds['X_col_name']
  x = SingleCellOMIC(X=ds['X'],
                     obs={'cellid': ds['X_row']},
                     var=var,
                     name=dataset_name)

  if 'y' in ds:
    var = {'protid': ds['y_col']}
    if 'y_col_name' in ds:
      var['protname'] = ds['y_col_name']
    y = SingleCellOMIC(X=ds['y'],
                       obs={'cellid': ds['X_row']},
                       var=var,
                       name=dataset_name)
  else:
    y = None
  return x, y


def get_scvi_dataset(dataset_name):
  """ Convert any SISUA dataset to relevant format for scVI models """
  from scvi.dataset import GeneExpressionDataset
  ds, gene, prot = get_dataset(dataset_name, override=False)
  X = np.concatenate((gene.X_train, gene.X_test), axis=0)
  labels = np.concatenate((prot.X_train, prot.X_test), axis=0)
  means, vars = get_library_size(X)
  is_multi_classes_labels = np.all(np.sum(labels, axis=1) != 1.)
  scvi = GeneExpressionDataset(X=X,
                               local_means=means,
                               local_vars=vars,
                               batch_indices=np.zeros(shape=(X.shape[0], 1)),
                               labels=None,
                               gene_names=gene.X_col,
                               cell_types=None)
  if not is_multi_classes_labels:
    scvi.labels = labels
    scvi.cell_types = prot.X_col
  else:
    scvi.labels = labels
    scvi.adt_expression = labels
    scvi.protein_markers = prot.X_col
  return scvi



from __future__ import print_function, division, absolute_import

import io
import os
from collections import defaultdict

import numpy as np

from odin.utils import stdio, get_formatted_datetime

from .inference_base import Inference
from ._consts import PREDICTION_BATCH_SIZE
from sisua.data import apply_artificial_corruption

# ===========================================================================
# Helper for DCA
# ===========================================================================
def _no_zero_genes(X):
  # check if any genes is all zeros (i.e. zero columns)
  X = X.astype('float32')
  zero_genes = np.sum(X, axis=0, keepdims=True)
  if np.sum(zero_genes == 0) > 0:
    X = X + np.finfo(X.dtype).eps
  return X

def dca_normalize(X, log=True):
  from dca.api import normalize
  from anndata import AnnData
  from odin.utils import catch_warnings_error
  # this is the default configuration
  log1p = bool(log)
  X = _no_zero_genes(X)
  with catch_warnings_error(Warning):
    adata = normalize(AnnData(X),
                      filter_min_counts=False,
                      size_factors=True,
                      normalize_input=True,
                      logtrans_input=log1p)
  return adata.X, adata.obs.size_factors.values

def dca_train(
    X, network, log, learning_rate,
    epochs=300, reduce_lr=10,
    early_stop=15, batch_size=32,
    validation_split=0.1, verbose=True):
  import keras.optimizers as opt
  from keras.callbacks import (EarlyStopping, ReduceLROnPlateau)
  from dca.api import normalize
  from anndata import AnnData
  adata = normalize(AnnData(X),
                    filter_min_counts=False,
                    size_factors=True,
                    normalize_input=True,
                    logtrans_input=log)

  model = network.model
  loss = network.loss

  optimizer = opt.__dict__['adam'](lr=learning_rate, clipvalue=5.)
  model.compile(loss=loss, optimizer=optimizer)

  # Callbacks
  callbacks = []
  if reduce_lr:
    lr_cb = ReduceLROnPlateau(monitor='val_loss', patience=reduce_lr, verbose=verbose)
    callbacks.append(lr_cb)
  if early_stop:
    es_cb = EarlyStopping(monitor='val_loss', patience=early_stop, verbose=verbose)
    callbacks.append(es_cb)

  inputs = {'count': adata.X, 'size_factors': adata.obs.size_factors}
  output = adata.raw.X

  loss = model.fit(inputs, output,
                   epochs=epochs,
                   batch_size=batch_size,
                   shuffle=True,
                   callbacks=callbacks,
                   validation_split=validation_split,
                   verbose=verbose)
  return loss

# ===========================================================================
# Main
# ===========================================================================
class InferenceDCA(Inference):
  """ InferenceDCA """

  def __init__(self, gene_dim,
               dispersion='gene-cell',
               xnorm='log', tnorm='raw', ynorm='prob',
               xclip=0, yclip=0,
               xdist='zinb', ydist='bernoulli', zdist='normal',
               xdrop=0.3, edrop=0, zdrop=0, ddrop=0,
               hdim=128, zdim=32, nlayer=2,
               batchnorm=True, analytic=True,
               kl_weight=1., warmup=400, y_weight=10.,
               extra_module_path=None,
               **kwargs):
    try:
      import dca as _dca_
      from keras import backend
    except ImportError as e:
      raise RuntimeError("pip install dca")
    # ====== validate the arguments ====== #
    model = kwargs.get('model', 'vae')
    prot_dim = None
    if model not in ('vae', 'dca'):
      raise RuntimeError("InferenceSCVI only support 'vae' model")
    if xdist not in ('zinb', 'nb'):
      raise ValueError("Only support 'zinb' and 'nb' for `xdist`")
    if xnorm not in ('log', 'raw'):
      raise ValueError(
          "Only support log variational or raw count variational for scVI")
    if tnorm not in ('raw',):
      raise ValueError(
          "Only support raw count target value for scVI")
    self._name = 'dca'
    # ====== others ====== #
    self._history = {'train': defaultdict(list),
                     'valid': defaultdict(list)}
    self._training_log = ''
    self._configs = dict(locals())
    self._configs.update(kwargs)

    self._gmm_threshold = {}
    self._gene_dim = int(gene_dim)
    self._prot_dim = 0 if prot_dim is None else max(int(prot_dim), 0)
    self._is_fitted = False

    self._ps = 0
    self._trained_n_epoch = 0
    self._batch_size = 0
    self._learning_rate = 0
    self._n_mcmc_train = 0
    self._corruption_rate = None
    self._corruption_dist = None
    # ====== initialize ====== #
    self._initialize_model()

  def _initialize_model(self, weights=None):
    #
    if self.xdist == 'zinb':
      if self.dispersion == 'gene':
        from dca.network import ZINBConstantDispAutoencoder as Autoencoder
      elif self.dispersion == 'gene-cell':
        from dca.network import ZINBAutoencoder as Autoencoder
      else:
        raise ValueError("No support for dispersion: %s" % self.dispersion)
    #
    elif self.xdist == 'nb':
      if self.dispersion == 'gene':
        from dca.network import NBConstantDispAutoencoder as Autoencoder
      elif self.dispersion == 'gene-cell':
        from dca.network import NBAutoencoder as Autoencoder
      else:
        raise ValueError("No support for dispersion: %s" % self.dispersion)
    #
    else:
      raise ValueError("No support for xdist: %s" % self.xdist)

    self._ae = Autoencoder(
        input_size=self.gene_dim,
        output_size=None,
        hidden_size=(self.hdim,) * self.nlayer + (self.zdim,) + (self.hdim,) * self.nlayer,
        input_dropout=self.xdrop,
        hidden_dropout=self.zdrop,
        batchnorm=True,
        activation='relu')
    self._ae.build()
    if weights is not None:
      self._ae.model.set_weights(weights)

  # ******************** pickling ******************** #
  def __getstate__(self):
    weights = self._ae.model.get_weights()
    states = self.get_states()
    return states, weights

  def __setstate__(self, states):
    states, weights = states
    self.set_states(states)
    self._initialize_model(weights)

  # ******************** fitting ******************** #
  def fit(self, X, y=None,
          supervised_percent=0.8, validation_percent=0.1,
          n_mcmc_samples=1,
          corruption_rate=0.25, corruption_dist='binomial',
          batch_size=64, n_epoch=120, learning_rate=1e-4,
          monitoring=False, fig_dir=None,
          detail_logging=False):
    training_log = io.StringIO()
    stdio(training_log)

    if self._corruption_rate is None:
      self._corruption_rate = corruption_rate
      self._corruption_dist = corruption_dist
    else:
      assert (self._corruption_rate == corruption_rate and
              self._corruption_dist == corruption_dist),\
      ("Model have been fitted with following corruption configurations: %s, %s;"
       "but given: %s, %s" % (self._corruption_rate, self._corruption_dist,
                              corruption_rate, corruption_dist))
    assert n_mcmc_samples == 1, \
    "Only support 1 MCMC sample during training for scVI"

    self._ps = supervised_percent
    self._batch_size = batch_size
    self._learning_rate = learning_rate
    self._n_mcmc_train = n_mcmc_samples
    X = apply_artificial_corruption(
        X, dropout=corruption_rate, distribution=corruption_dist)
    X = _no_zero_genes(X)
    # ====== training ====== #
    hist = dca_train(
        X=X, network=self._ae, log=self.xnorm == 'log',
        learning_rate=learning_rate, epochs=n_epoch, batch_size=batch_size,
        validation_split=validation_percent, verbose=detail_logging)
    # ====== save history ====== #
    hist = hist.history
    self._history['train']['loss'] += hist["loss"]
    self._history['valid']['loss'] += hist["val_loss"]
    # ====== update states ====== #
    self._trained_n_epoch += n_epoch
    self._is_fitted = True
    stdio()
    training_log.seek(0)
    self._training_log += '\n%s\n' % get_formatted_datetime(only_number=False)
    self._training_log += training_log.read()
    training_log.close()

  # ******************** history ******************** #
  def get_train_loss(self):
    return self._history['train']['loss']

  def get_valid_loss(self):
    return self._history['valid']['loss']

  def get_train_history(self, name):
    return self._history['train'][name]

  def get_valid_history(self, name):
    return self._history['valid'][name]

  # ******************** scoring ******************** #
  def score(self, X, y=None, n_mcmc_samples=100, return_mean=True):
    """Compute the per-sample average of recorded metrics of the given data X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_dimensions)
        List of n_features-dimensional data points. Each row
        corresponds to a single data point.

    Returns
    -------
    dictionary : metric_name -> values
    """
    # only VAE model for scVI
    # X, _, L, _, _, _ = self._preprocess_inputs(X, y=None)
    # L = np.exp(L)

    X, L = dca_normalize(X, log=self.xnorm == 'log')
    loss = self._ae.model.evaluate({
        'count': X,
        'size_factors': L}, y=X, batch_size=PREDICTION_BATCH_SIZE,
        verbose=0)
    return {'loss': loss}

  # ******************** predicting ******************** #
  def predict_Z(self, X):
    """ Return mean of the latent posterior
    (i.e. mean of Normal distribution) """
    X, L = dca_normalize(X, log=self.xnorm == 'log')
    return self._ae.encoder.predict({
        'count': X,
        'size_factors': L}, batch_size=PREDICTION_BATCH_SIZE)

  def predict_y(self, X, y, n_mcmc_samples=100):
    """ Predicting the protein marker """
    assert self.is_semi_supervised and y is not None,\
    "Only semi-supervised model has this prediction, and `y` must not be None"

  def predict_L(self, X, n_mcmc_samples=100):
    """ Return library size prediction """
    v = self.predict_V(X, n_mcmc_samples=n_mcmc_samples)
    return np.sum(v, axis=1, keepdims=True)

  def predict_Lstddev(self, X, n_mcmc_samples=100):
    """ Return library size stddev prediction """
    v_stddev = self.predict_Vstddev(X, n_mcmc_samples)
    return np.sum(v_stddev, axis=1, keepdims=True)

  def predict_W(self, X, n_mcmc_samples=100):
    """ Return mean of the reconstructed value """
    v = self.predict_V(X, n_mcmc_samples)
    if 'pi' not in self._ae.extra_models:
      return v
    pi = self.predict_PI(X, n_mcmc_samples)
    return (1 - pi) * v

  def predict_Wstddev(self, X, n_mcmc_samples=100):
    """ Return stddev of reconstructed value
    if not a variational model, then V_stddev is None
    """
    v_stddev = self.predict_Vstddev(X, n_mcmc_samples)
    if 'pi' not in self._ae.extra_models:
      return v_stddev
    v = self.predict_V(X, n_mcmc_samples)
    pi = self.predict_PI(X, n_mcmc_samples)
    w_mean = (1 - pi) * v
    return np.sqrt(
        (1 - pi) * (v_stddev + v**2) - w_mean**2)

  def predict_V(self, X, n_mcmc_samples = 100):
    """ Return mean of the denoised value """
    X, L = dca_normalize(X, log=self.xnorm == 'log')
    return self._ae.model.predict({
        'count': X,
        'size_factors': L}, batch_size=PREDICTION_BATCH_SIZE)

  def predict_Vstddev(self, X, n_mcmc_samples = 100):
    """ Return stddev of denoised value
    if not a variational model, then V_stddev is None
    """
    X, L = dca_normalize(X, log=self.xnorm == 'log')
    return self._ae.extra_models['dispersion'].predict({
        'count': X,
        'size_factors': L}, batch_size=PREDICTION_BATCH_SIZE)

  def predict_PI(self, X, n_mcmc_samples=100):
    """ Return a matrix (n_sample, n_gene) of Zero-inflated
    rate

    if not a zero-inflated model, then return None
    """
    if 'pi' not in self._ae.extra_models:
      raise RuntimeError("This is not zero-inflated model.")
    X, L = dca_normalize(X, log=self.xnorm == 'log')
    return self._ae.extra_models['pi'].predict({
        'count': X,
        'size_factors': L}, batch_size=PREDICTION_BATCH_SIZE)

  # ******************** sampling methods ******************** #
  def sample_Z(self, X, n_mcmc_samples=1):
    raise NotImplementedError

  def sample_y(self, X, n_mcmc_samples=1):
    raise NotImplementedError

  def sample_W(self, X, n_mcmc_samples=1):
    raise NotImplementedError

  def sample_V(self, X, n_mcmc_samples=1):
    raise NotImplementedError

  def sample_L(self, X, n_mcmc_samples=1):
    raise NotImplementedError



from __future__ import absolute_import, division, print_function

import os
from collections import OrderedDict

import numpy as np
import pandas as pd
import seaborn as sbn
import tensorflow as tf
from sklearn.metrics import (confusion_matrix, explained_variance_score,
                             f1_score, mean_absolute_error, mean_squared_error,
                             r2_score)

from odin import backend as K
from odin.ml import fast_pca, fast_tsne
from odin.utils import async_mpi, catch_warnings_ignore, ctext, one_hot
from odin.visual import (generate_random_colors, generate_random_marker,
                         plot_confusion_matrix, plot_figure, plot_gridSpec,
                         plot_gridSubplot, plot_histogram, plot_save,
                         plot_scatter, plot_scatter_heatmap,
                         plot_scatter_layers, plot_series_statistics, subplot)
from sisua.utils.others import anything2image


# ===========================================================================
# Some helper
# ===========================================================================
def _clip_count_sum(x):
  return np.clip(x,
                 a_min=np.min(x),
                 a_max=np.percentile(x, q=95))

def _clipping_quartile(x, alpha=1.5):
  # result wider range threshold, keep more data points,
  # lower LLK
  x = x.astype('int32')
  Q1 = np.percentile(x, q=25)
  Q3 = np.percentile(x, q=75)
  IQR = Q3 - Q1
  x = x[x < Q3 + alpha * IQR]
  x = x[x > Q1 - alpha * IQR]
  return x

def downsample_data(*X):
  y = [None] * len(X)
  _ = list(set(x.shape[0] for x in X
               if x is not None))
  assert len(_) == 1, "Inconsistent shape[0] for X and y"
  num_samples = _[0]
  _RAND = np.random.RandomState(seed=87654321)
  # ====== Downsample if the data is huge ====== #
  if num_samples > 8000:
    print("[Warning] Given: %s; downsample to 8000 samples" %
      ctext(', '.join([str(x.shape) for x in X
                      if x is not None]), 'cyan'))
    ids = _RAND.choice(a=np.arange(0, num_samples),
                       size=8000, replace=False)
    for i, x in enumerate(X):
      if x is not None:
        x = x[ids]
      y[i] = x
  else:
    y = X
  return tuple(y)

def show_image(x, is_probability=False):
  from matplotlib import pyplot as plt
  from odin import backend as K
  x = anything2image(x)

  if x.shape[0] > 32:
    x = tf.nn.max_pool(np.reshape(x, (1, x.shape[0], x.shape[1], 1)),
                       ksize=(1, 4, 4, 1),
                       strides=(1, 4, 4, 1),
                       padding="SAME")
    x = np.squeeze(K.eval(x))
  ax = plt.gca()
  plt.imshow(x, interpolation='nearest', cmap=plt.cm.Greys_r,
             vmin=0. if is_probability else None,
             vmax=1. if is_probability else None)
  plt.xticks(()); ax.set_xticklabels([])
  plt.yticks(()); ax.set_yticklabels([])
  ax.set_aspect(aspect='auto')

def fast_scatter(x, y, labels, title,
                 azim=None, elev=None, ax=None,
                 enable_legend=False,
                 size=18, fontsize=12):
  y = np.squeeze(y)
  if y.ndim == 1:
    pass
  elif y.ndim == 2: # provided one-hot vectors
    y = np.argmax(y, axis=-1)
  else:
    raise ValueError("No support for `y` shape: %s" % str(y.shape))
  # ====== get colors and legends ====== #
  if labels is not None:
    y = [labels[int(i)] for i in y]
    num_classes = len(labels)
  else:
    num_classes = len(np.unique(y))
  # int(np.ceil(num_classes / 2)) if num_classes <= 20 else num_classes // 5
  plot_scatter(x=x, color=y, marker=y,
      size=size, azim=azim, elev=elev,
      legend_enable=enable_legend,
      legend_ncol=3,
      fontsize=fontsize, ax=ax, title=title)

# ===========================================================================
# IO
# ===========================================================================
def save_figures(figures, path, dpi=None,
                 separate_files=True, clear_figures=True):
  if os.path.isfile(path) or '.pdf' == path[-4:].lower():
    separate_files = False
    assert '.pdf' == path[-4:].lower(), \
    "If a file is given, it must be PDF file"
  assert isinstance(figures, dict), \
  "figures must be dictionary mapping from figure name to matplotlib.Figure"
  n_figures = len(figures)
  # ====== saving PDF file ====== #
  if not separate_files:
    if dpi is None:
      dpi = 48

    if '.pdf' not in path:
      path = path + '.pdf'
    from matplotlib.backends.backend_pdf import PdfPages
    pp = PdfPages(path)
    for key, fig in figures.items():
      try:
        fig.savefig(pp, dpi=dpi, format='pdf', bbox_inches="tight")
      except Exception as e:
        print("Error:", key)
        print(" ", e)
    pp.close()
  # ====== saving PNG file ====== #
  else:
    if dpi is None:
      dpi = 160

    if not os.path.exists(path):
      os.mkdir(path)
    assert os.path.isdir(path), "'%s' must be path to a folder" % path
    kwargs = dict(dpi=dpi, bbox_inches="tight")
    for key, fig in figures.items():
      out_path = os.path.join(path, key + '.png')
      try:
        fig.savefig(out_path, **kwargs)
      except Exception as e:
        print("Error:", ctext(out_path, 'red'))
        print(" ", e)
  # ====== clear figures ====== #
  print("%s figures save to path: %s" %
    (ctext(n_figures, 'lightcyan'), ctext(path, 'lightyellow')))
  if clear_figures:
    figures.clear()

# ===========================================================================
# Evaluating the reconstruction
# ===========================================================================
def plot_evaluate_reconstruction(X, W, y_raw, y_prob,
                                 X_row, X_col, labels, title,
                                 pi=None,
                                 enable_image=True,
                                 enable_tsne=True,
                                 enable_sparsity=True):
  """
  pi : zero-inflated rate (imputation rate), or dropout probabilities
  """
  print("Evaluate: [Reconstruction]", ctext(title, 'lightyellow'))
  from matplotlib import pyplot as plt
  fontsize = 12

  W_stdev_total, W_stdev_explained = None, None
  if isinstance(W, (tuple, list)):
    if len(W) == 1:
      W = W[0]
    elif len(W) == 3:
      W, W_stdev_total, W_stdev_explained = W
    else:
      raise RuntimeError()
  elif W.ndim == 3:
    W, W_stdev_total, W_stdev_explained = W[0], W[1], W[2]
  # convert the prediction to integer
  # W = W.astype('int32')

  assert (X.shape[0] == W.shape[0] == y_raw.shape[0]) and \
  (X.shape == W.shape) and \
  (y_raw.shape == y_prob.shape)

  X, X_row, W, y_raw, y_prob, pi = downsample_data(X, X_row, W, y_raw, y_prob, pi)
  y_argmax = np.argmax(y_prob, axis=-1)

  # ====== prepare count-sum ====== #
  X_log = K.log_norm(X, axis=1)
  W_log = K.log_norm(W, axis=1)

  X_gene_countsum = np.sum(X, axis=0)
  X_cell_countsum = np.sum(X, axis=1)
  X_gene_nzeros = np.sum(X == 0, axis=0)
  X_cell_nzeros = np.sum(X == 0, axis=1)

  gene_sort = np.argsort(X_gene_countsum)
  cell_sort = np.argsort(X_cell_countsum)

  W_gene_countsum = np.sum(W, axis=0)
  W_cell_countsum = np.sum(W, axis=1)
  W_gene_nzeros = np.sum(W == 0, axis=0)
  W_cell_nzeros = np.sum(W == 0, axis=1)

  X_col_sorted = X_col[gene_sort] if X_col is not None else None
  X_row_sorted = X_row[cell_sort] if X_row is not None else None

  if pi is not None:
    pi_cell_countsum = np.mean(pi, axis=1)
    pi_gene_countsum = np.mean(pi, axis=0)

  # ====== Compare image ====== #
  if enable_image:
    _RAND = np.random.RandomState(seed=87654321)
    n_img = 12
    n_img_row = min(3, X.shape[0] // n_img)
    n_row_per_row = 2 if pi is None else 3
    plot_figure(nrow=n_img_row * 4, ncol=18)
    count = 1
    all_ids = _RAND.choice(np.arange(0, X.shape[0]),
                       size=n_img * n_img_row,
                       replace=False)

    for img_row in range(n_img_row):
      ids = all_ids[img_row * n_img: (img_row + 1) * n_img]

      # plot original images
      for _, i in enumerate(ids):
        ax = plt.subplot(n_row_per_row * n_img_row, n_img, count)
        show_image(X[i])
        if _ == 0:
          plt.ylabel("Original")
        if X_row is not None:
          ax.set_title(X_row[i], fontsize=8)
        count += 1
      # plot reconstructed images
      for _, i in enumerate(ids):
        plt.subplot(n_row_per_row * n_img_row, n_img, count)
        show_image(W[i])
        if _ == 0:
          plt.ylabel("Reconstructed")
        count += 1
      # plot zero-inflated rate
      if pi is not None:
        for _, i in enumerate(ids):
          plt.subplot(n_row_per_row * n_img_row, n_img, count)
          show_image(pi[i], is_probability=True)
          if _ == 0:
            plt.ylabel("$p_{zero-inflated}$")
          count += 1
    plt.tight_layout()
  # ====== compare the T-SNE plot ====== #
  if enable_tsne:
    def pca_and_tsne(x, w):
      x_pca, w_pca = fast_pca(x, w, n_components=512,
                              random_state=87654321)
      x_tsne = fast_tsne(x_pca, n_components=2, random_state=87654321)
      w_tsne = fast_tsne(w_pca, n_components=2, random_state=87654321)
      return x_pca[:, :2], x_tsne, w_pca[:, :2], w_tsne
    # transforming the data
    (X_cell_pca, X_cell_tsne,
     W_cell_pca, W_cell_tsne) = pca_and_tsne(X_log, W_log)
    (X_gene_pca, X_gene_tsne,
     W_gene_pca, W_gene_tsne) = pca_and_tsne(X_log.T, W_log.T)
    # prepare the figure
    n_plot = 3 + 2 # 3 for cells, 2 for genes
    if pi is not None:
      n_plot += 2 # 2 more row for pi
    plot_figure(nrow=n_plot * 5, ncol=18)
    # Cells
    fast_scatter(x=X_cell_pca, y=y_argmax, labels=labels,
                 title="[PCA]Original Cell Data",
                 ax=(n_plot, 2, 1), enable_legend=False)
    fast_scatter(x=W_cell_pca, y=y_argmax, labels=labels,
                 title="[PCA]Reconstructed Cell Data",
                 ax=(n_plot, 2, 2), enable_legend=False)

    fast_scatter(x=X_cell_tsne, y=y_argmax, labels=labels,
                 title="[t-SNE]Original Cell Data",
                 ax=(n_plot, 2, 3), enable_legend=True)
    fast_scatter(x=W_cell_tsne, y=y_argmax, labels=labels,
                 title="[t-SNE]Reconstructed Cell Data",
                 ax=(n_plot, 2, 4), enable_legend=False)

    fast_log = lambda x: K.log_norm(x, axis=0)

    plot_scatter_heatmap(x=X_cell_tsne, val=fast_log(X_cell_countsum),
                         title="[t-SNE]Original Cell Data + Original Cell Countsum",
                         ax=(n_plot, 2, 5), colorbar=True)
    plot_scatter_heatmap(x=X_cell_tsne, val=fast_log(W_cell_countsum),
                         title="[t-SNE]Original Cell Data + Reconstructed Cell Countsum",
                         ax=(n_plot, 2, 6), colorbar=True)
    # Genes
    plot_scatter_heatmap(x=X_gene_pca, val=fast_log(X_gene_countsum),
                         title="[PCA]Original Gene Data + Original Gene Countsum",
                         ax=(n_plot, 2, 7), colorbar=True)
    plot_scatter_heatmap(x=W_gene_pca, val=fast_log(X_gene_countsum),
                         title="[PCA]Reconstructed Gene Data + Original Gene Countsum",
                         ax=(n_plot, 2, 8), colorbar=True)

    plot_scatter_heatmap(x=X_gene_tsne, val=fast_log(X_gene_countsum),
                         title="[t-SNE]Original Gene Data + Original Gene Countsum",
                         ax=(n_plot, 2, 9), colorbar=True)
    plot_scatter_heatmap(x=X_gene_tsne, val=fast_log(W_gene_countsum),
                         title="[t-SNE]Original Gene Data + Reconstructed Gene Countsum",
                         ax=(n_plot, 2, 10), colorbar=True)
    # zero-inflation rate
    if pi is not None:
      plot_scatter_heatmap(x=X_cell_tsne, val=X_cell_countsum,
                           title="[t-SNE]Original Cell Data + Original Cell Countsum",
                           ax=(n_plot, 2, 11), colorbar=True)
      plot_scatter_heatmap(x=X_cell_tsne, val=pi_cell_countsum,
                           title="[t-SNE]Original Cell Data + Zero-inflated rate",
                           ax=(n_plot, 2, 12), colorbar=True)

      plot_scatter_heatmap(x=X_gene_tsne, val=X_gene_countsum,
                           title="[t-SNE]Original Gene Data + Original Gene Countsum",
                           ax=(n_plot, 2, 13), colorbar=True)
      plot_scatter_heatmap(x=X_gene_tsne, val=pi_gene_countsum,
                           title="[t-SNE]Original Gene Data + Zero-inflated rate",
                           ax=(n_plot, 2, 14), colorbar=True)
    plt.tight_layout()
  # ******************** sparsity ******************** #
  if enable_sparsity:
    plot_figure(nrow=8, ncol=8)
    # ====== sparsity ====== #
    z = (X.ravel() == 0).astype('int32')
    z_res = (W.ravel() == 0).astype('int32')
    plot_confusion_matrix(ax=None,
        cm=confusion_matrix(y_true=z, y_pred=z_res, labels=(0, 1)),
        labels=('Not Zero', 'Zero'), colorbar=True, fontsize=fontsize + 4,
        title="Sparsity")

# ===========================================================================
# Streamline classifier
# ===========================================================================
def plot_evaluate_classifier(y_pred, y_true, labels, title,
                             show_plot=True, return_figure=False):
  """ Return a dictionary of scores
  {
      F1micro=f1_micro * 100,
      F1macro=f1_macro * 100,
      F1weight=f1_weight * 100,
      F1_[classname]=...
  }
  """
  from matplotlib import pyplot as plt
  fontsize = 12
  num_classes = len(labels)
  nrow = int(np.ceil(num_classes / 5))
  ncol = int(np.ceil(num_classes / nrow))

  if y_pred.ndim == 1:
    y_pred = one_hot(y_pred, nb_classes=num_classes)
  if y_true.ndim == 1:
    y_true = one_hot(y_true, nb_classes=num_classes)

  if show_plot:
    fig = plot_figure(nrow=4 * nrow + 2, ncol=4 * ncol)

  f1_classes = []
  for i, (name, pred, true) in enumerate(zip(labels, y_pred.T, y_true.T)):
    f1_classes.append(f1_score(true, pred))
    if show_plot:
      plot_confusion_matrix(confusion_matrix(y_true=true, y_pred=pred),
                            labels=[0, 1],
                            fontsize=fontsize,
                            ax=(nrow, ncol, i + 1),
                            title=name + '\n')

  f1_micro = f1_score(y_true=y_true.ravel(), y_pred=y_pred.ravel())
  f1_macro = np.mean(f1_classes)
  f1_weight = f1_score(y_true=y_true, y_pred=y_pred, average='weighted')

  if show_plot:
    plt.suptitle('%s\nF1-micro:%.2f  F1-macro:%.2f  F1-weight:%.2f' %
                 (title, f1_micro * 100, f1_macro * 100, f1_weight * 100))
    plt.tight_layout(rect=[0, 0.03, 1, 0.96])

  results = dict(
      F1micro=f1_micro * 100,
      F1macro=f1_macro * 100,
      F1weight=f1_weight * 100,
  )
  for name, f1 in zip(labels, f1_classes):
    results['F1_' + name] = f1 * 100

  if show_plot and return_figure:
    return results, fig
  return results

def plot_evaluate_regressor(y_pred, y_true, labels, title):
  from matplotlib import pyplot as plt
  num_classes = len(labels)
  nbins = 120
  fontsize = 8
  y_pred = np.round(y_pred).astype('int32')
  y_true = y_true.astype('int32')
  # ====== basic scores ====== #
  r2 = r2_score(y_true, y_pred)
  var = explained_variance_score(y_true, y_pred)
  mse = mean_squared_error(y_true, y_pred)
  mae = mean_absolute_error(y_true, y_pred)

  # ====== helper ====== #
  def plot_hist(hist, ax, name):
    count, bins = plot_histogram(true, bins=nbins, ax=ax, title=name, fontsize=fontsize)
    plt.xlim((np.min(bins), np.max(bins)))
    plt.xticks(np.linspace(start=np.min(bins), stop=np.max(bins), num=8, dtype='int32'),
               fontsize=6)
    plt.yticks(np.linspace(start=np.min(count), stop=np.max(count), num=8, dtype='int32'),
               fontsize=6)

  # ====== raw count prediction ====== #
  plot_figure(nrow=4, ncol=num_classes * 2)
  for i in range(num_classes):
    name = labels[i]
    r2_ = r2_score(y_true=y_true[:, i],
                   y_pred=y_pred[:, i])
    pred = _clipping_quartile(y_pred[:, i])
    true = _clipping_quartile(y_true[:, i])
    plot_hist(hist=true, ax=(2, num_classes, i + 1),
              name='[%s] R2: %.6f' % (name, r2_))
    if i == 0:
      plt.ylabel('True')
    plot_hist(hist=pred, ax=(2, num_classes, num_classes + i + 1), name=None)
    if i == 0:
      plt.ylabel('Pred')
  # set the title
  plt.suptitle('[%s]  R2: %.6f   ExpVAR: %.6f   MSE: %.6f   MAE: %.6f' %
    (str(title), r2, var, mse, mae), fontsize=fontsize + 1)

# ===========================================================================
# Monitoring the training
# ===========================================================================
def _show_zero_inflated_pi(p, ax, handles, indices):
  ax = ax.twinx()
  _, = ax.plot(p[indices],
               marker='o', linestyle='None',
               markersize=0.8, color='blue',
               alpha=0.3, label="$p_{dropout}$")
  ax.set_ylabel("Zero-inflation rate", fontsize=8)
  ax.tick_params(axis='both', labelsize=8)
  handles.append(_)

@async_mpi
def plot_monitoring_epoch(X, X_drop, y, Z, Z_drop,
                          W_outputs, W_drop_outputs,
                          pi, pi_drop,
                          row_name, dropout_percentage,
                          curr_epoch, ds_name, labels, save_dir):
  # Order of W_outputs: [W, W_stdev_total, W_stdev_explained]
  from matplotlib import pyplot as plt
  if y.ndim == 2:
    y = np.argmax(y, axis=-1)
  y = np.array([labels[i] for i in y])
  dropout_percentage_text = '%g%%' % (dropout_percentage * 100)

  Z_pca = fast_pca(Z, n_components=2, random_state=5218)
  Z_pca_drop = fast_pca(Z_drop, n_components=2, random_state=5218)
  if W_outputs is not None:
    X_pca, X_pca_drop, W_pca, W_pca_drop = fast_pca(
        X, X_drop, W_outputs[0], W_drop_outputs[0],
        n_components=2, random_state=5218)
   # ====== downsampling ====== #
  rand = np.random.RandomState(seed=5218)
  n_test_samples = len(y)
  ids = np.arange(n_test_samples, dtype='int32')
  if n_test_samples > 8000:
    ids = rand.choice(ids, size=8000, replace=False)
  # ====== scatter configuration ====== #
  config = dict(size=6, labels=None)
  y = y[ids]

  X = X[ids]
  X_drop = X_drop[ids]
  Z_pca = Z_pca[ids]
  X_pca = X_pca[ids]
  W_pca = W_pca[ids]

  W_outputs = [w[ids] for w in W_outputs]
  W_drop_outputs = [w[ids] for w in W_drop_outputs]
  Z_pca_drop = Z_pca_drop[ids]
  X_pca_drop = X_pca_drop[ids]
  W_pca_drop = W_pca_drop[ids]

  if pi is not None:
    pi = pi[ids]
    pi_drop = pi_drop[ids]
  # ====== plotting NO reconstruction ====== #
  if W_outputs is None:
    plot_figure(nrow=8, ncol=20)
    fast_scatter(x=Z_pca, y=y,
                 title="[PCA] Test data latent space",
                 enable_legend=True, ax=(1, 2, 1), **config)
    fast_scatter(x=Z_pca_drop, y=y,
                 title="[PCA][Dropped:%s] Test data latent space" % dropout_percentage_text,
                 ax=(1, 2, 2), **config)
  # ====== plotting WITH reconstruction ====== #
  else:
    plot_figure(nrow=16, ncol=20)
    # original test data WITHOUT dropout
    fast_scatter(x=X_pca, y=y, title="[PCA][Test Data] Original",
      ax=(2, 3, 1), **config)
    fast_scatter(x=W_pca, y=y, title="Reconstructed",
      ax=(2, 3, 2), **config)
    fast_scatter(x=Z_pca, y=y, title="Latent space",
      ax=(2, 3, 3), **config)
    # original test data WITH dropout
    fast_scatter(x=X_pca_drop, y=y,
      title="[PCA][Dropped:%s][Test Data] Original" % dropout_percentage_text,
      ax=(2, 3, 4), **config)
    fast_scatter(x=W_pca_drop, y=y, title="Reconstructed",
      ax=(2, 3, 5), enable_legend=True, **config)
    fast_scatter(x=Z_pca_drop, y=y, title="Latent space",
      ax=(2, 3, 6), **config)
  plot_save(os.path.join(save_dir, 'latent_epoch%d.png') % curr_epoch,
            dpi=180, clear_all=True, log=True)
  # ====== plot count-sum ====== #
  if W_outputs is not None:
    X_countsum = _clip_count_sum(np.sum(X, axis=-1))
    W_countsum = _clip_count_sum(np.sum(W_outputs[0], axis=-1))
    X_drop_countsum = _clip_count_sum(np.sum(X_drop, axis=-1))
    W_drop_countsum = _clip_count_sum(np.sum(W_drop_outputs[0], axis=-1))
    series_config = [dict(xscale='linear', yscale='linear', sort_by=None),
                     dict(xscale='linear', yscale='linear', sort_by='expected')]

    if pi is not None:
      pi_sum = np.mean(pi, axis=-1)
      pi_drop_sum = np.mean(pi_drop, axis=-1)
    # plot the reconstruction count sum
    plot_figure(nrow=3 * 5 + 8, ncol=18)
    with plot_gridSpec(nrow=3 * (2 if pi is None else 3) + 4 * 3 + 1,
                       ncol=6,
                       wspace=1.0, hspace=0.8) as grid:
      kws = dict(colorbar=True, fontsize=10, size=10, marker=y, n_samples=1200)
      # without dropout
      ax = subplot(grid[:3, 0:3])
      plot_scatter_heatmap(x=X_pca, val=X_countsum, ax=ax,
                           legend_enable=False, title='Original data (Count-sum)',
                           **kws)
      ax = subplot(grid[:3, 3:6])
      plot_scatter_heatmap(x=W_pca, val=W_countsum, ax=ax,
                           legend_enable=False, title='Reconstruction (Count-sum)',
                           **kws)
      # with dropout
      ax = subplot(grid[3:6, 0:3])
      plot_scatter_heatmap(x=X_pca_drop, val=X_drop_countsum, ax=ax,
                           legend_enable=True if pi is None else False,
                           legend_ncol=len(labels),
                           title='[Dropped:%s]Original data (Count-sum)' % dropout_percentage_text,
                           **kws)
      ax = subplot(grid[3:6, 3:6])
      plot_scatter_heatmap(x=W_pca_drop, val=W_drop_countsum, ax=ax,
                           legend_enable=False,
                           title='[Dropped:%s]Reconstruction (Count-sum)' % dropout_percentage_text,
                           **kws)
      row_start = 6
      # zero-inflated pi
      if pi is not None:
        ax = subplot(grid[6:9, 0:3])
        plot_scatter_heatmap(x=X_pca, val=pi_sum, ax=ax,
                             legend_enable=True, legend_ncol=len(labels),
                             title='Zero-inflated probabilities',
                             **kws)
        ax = subplot(grid[6:9, 3:6])
        plot_scatter_heatmap(x=X_pca, val=pi_drop_sum, ax=ax,
                             legend_enable=False,
                             title='[Dropped:%s]Zero-inflated probabilities' % dropout_percentage_text,
                             **kws)
        row_start += 3

      # plot the count-sum series
      def plot_count_sum_series(x, w, p, row_start, tit):
        if len(w) != 3: # no statistics provided
          return
        expected, stdev_total, stdev_explained = w
        count_sum_observed = np.sum(x, axis=0)
        count_sum_expected = np.sum(expected, axis=0)
        count_sum_stdev_total = np.sum(stdev_total, axis=0)
        count_sum_stdev_explained = np.sum(stdev_explained, axis=0)
        if p is not None:
          p_sum = np.mean(p, axis=0)
        for i, kws in enumerate(series_config):
          ax = subplot(grid[row_start:row_start + 3, (i * 3):(i * 3 + 3)])
          ax, handles, indices = plot_series_statistics(
              count_sum_observed, count_sum_expected,
              explained_stdev=count_sum_stdev_explained,
              total_stdev=count_sum_stdev_total,
              fontsize=8, ax=ax, legend_enable=False,
              title=tit if i == 0 else None,
              despine=True if p is None else False,
              return_handles=True, return_indices=True,
              **kws)
          if p is not None:
            _show_zero_inflated_pi(p_sum, ax, handles, indices)
          plt.legend(handles=handles, loc='best', markerscale=4, fontsize=8)
      # add one row extra padding
      row_start += 1
      plot_count_sum_series(x=X, w=W_outputs, p=pi,
                            row_start=row_start,
                            tit="Count-sum X_original - W_original")
      row_start += 1
      plot_count_sum_series(x=X_drop, w=W_drop_outputs, p=pi_drop,
                            row_start=row_start + 3,
                            tit="[Dropped:%s]Count-sum X_drop - W_dropout" % dropout_percentage_text)
      row_start += 1
      plot_count_sum_series(x=X, w=W_drop_outputs, p=pi_drop,
                            row_start=row_start + 6,
                            tit="[Dropped:%s]Count-sum X_original - W_dropout" % dropout_percentage_text)
    plot_save(os.path.join(save_dir, 'countsum_epoch%d.png') % curr_epoch,
              dpi=180, clear_all=True, log=True)
  # ====== plot series of samples ====== #
  if W_outputs is not None and len(W_outputs) == 3:
    # NOTe: turn off pi here
    pi = None

    n_visual_samples = 8
    plot_figure(nrow=3 * n_visual_samples + 8, ncol=25)
    col_width = 5
    with plot_gridSpec(nrow=3 * n_visual_samples,
                       ncol=4 * col_width,
                       wspace=5.0, hspace=1.0) as grid:
      curr_grid_index = 0
      for i in rand.permutation(len(X))[:n_visual_samples]:
        observed = X[i]
        expected, stdev_explained, stdev_total = [w[i] for w in W_outputs]
        expected_drop, stdev_explained_drop, stdev_total_drop = [w[i] for w in W_drop_outputs]
        if pi is not None:
          p_zi = pi[i]
          p_zi_drop = pi_drop[i]
        # compare to W_original
        for j, kws in enumerate(series_config):
          ax = subplot(grid[curr_grid_index:curr_grid_index + 3,
                            (j * col_width):(j * col_width + col_width)])
          ax, handles, indices = plot_series_statistics(
              observed, expected,
              explained_stdev=stdev_explained,
              total_stdev=stdev_total,
              fontsize=8, legend_enable=False,
              despine=True if pi is None else False,
              title=("'%s' X_original - W_original" % row_name[i])
              if j == 0 else None,
              return_handles=True, return_indices=True,
              **kws)
          if pi is not None:
            _show_zero_inflated_pi(p_zi, ax, handles, indices)
          plt.legend(handles=handles, loc='best', markerscale=4, fontsize=8)
        # compare to W_dropout
        for j, kws in enumerate(series_config):
          col_start = col_width * 2
          ax = subplot(grid[curr_grid_index:curr_grid_index + 3,
                            (col_start + j * col_width):(col_start + j * col_width + col_width)])
          ax, handles, indices = plot_series_statistics(
              observed, expected_drop,
              explained_stdev=stdev_explained_drop,
              total_stdev=stdev_total_drop,
              fontsize=8, legend_enable=False,
              despine=True if pi is None else False,
              title=("[Dropped:%s]'%s' X_original - W_dropout" %
              (dropout_percentage_text, row_name[i])) if j == 0 else None,
              return_handles=True, return_indices=True,
              **kws)
          if pi is not None:
            _show_zero_inflated_pi(p_zi_drop, ax, handles, indices)
          plt.legend(handles=handles, loc='best', markerscale=4, fontsize=8)
        curr_grid_index += 3
    plot_save(os.path.join(save_dir, 'samples_epoch%d.png') % curr_epoch,
              dpi=180, clear_all=True, log=True)
  # ====== special case for mnist ====== #
  if 'mnist' in ds_name and W_outputs is not None:
    plot_figure(nrow=3, ncol=18)
    n_images = 32
    ids = rand.choice(np.arange(X.shape[0], dtype='int32'),
                      size=n_images, replace=False)
    meta_data = [
        ("Org", X[ids]),
        ("Rec", W_outputs[0][ids]),
        ("OrgDropout", X_drop[ids]),
        ("RecDropout", W_drop_outputs[0][ids])
    ]
    count = 1
    for name, data in meta_data:
      for i in range(n_images):
        x = data[i].reshape(28, 28)
        plt.subplot(4, n_images, count); show_image(x)
        if i == 0:
          plt.ylabel(name, fontsize=8)
        count += 1
    plt.subplots_adjust(wspace=0.05, hspace=0.05)
    plot_save(os.path.join(save_dir, 'image_epoch%d.png') % curr_epoch,
              dpi=180, clear_all=True, log=True)



# /data1/czi_data/GSE75478_velten
from __future__ import absolute_import, division, print_function

import base64
import gzip
import os
import pickle
import shutil
from io import BytesIO, StringIO

import numpy as np
import scipy as sp

from odin.fuel import Dataset
from odin.utils import crypto, ctext, get_file
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR
from sisua.data.utils import remove_allzeros_columns, save_to_dataset

_URL = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL0tJX0ZBQ1NfJWRwcm90ZWluLnpp\ncA==\n'
_FACS_PREPROCESSED = os.path.join(PREPROCESSED_BASE_DIR, 'FACS%d_preprocessed')

_PASSWORD = 'uef-czi'


def read_FACS(n_protein, override=False, verbose=False):
  download_path = os.path.join(DOWNLOAD_DIR, "FACS_original")
  if not os.path.exists(download_path):
    os.mkdir(download_path)

  n_protein = int(n_protein)
  assert n_protein in (2, 5)

  preprocessed_path = _FACS_PREPROCESSED % n_protein
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)
  elif override:
    shutil.rmtree(preprocessed_path)
    os.mkdir(preprocessed_path)
  # ******************** preprocessed data NOT found ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    X, X_row, X_col = [], None, None
    y, y_row, y_col = [], None, None
    # ====== download the data ====== #
    url = str(base64.decodebytes(_URL), 'utf-8') % n_protein
    base_name = os.path.basename(url)
    get_file(fname=base_name, origin=url, outdir=download_path, verbose=verbose)
    zip_path = os.path.join(download_path, base_name)
    # ====== extract the data ====== #
    data_dict = {}
    for name, data in crypto.unzip_aes(zip_path,
                                       password=_PASSWORD,
                                       verbose=False):
      base_name = os.path.splitext(name)[0]
      if '.npz' in name:
        data = sp.sparse.load_npz(BytesIO(data)).todense()
      elif '.csv' in name:
        data = np.loadtxt(StringIO(str(data, 'utf-8')),
                          dtype=str,
                          delimiter=',')
      else:
        raise RuntimeError("Unknown format: %s" % name)
      data_dict[base_name] = data
      if verbose:
        print('%-12s' % base_name, ':', data.shape)
    # ====== post-processing ====== #
    X = data_dict['X'].astype('float32')
    X = np.array(X)
    X_row, X_col = data_dict['X_row'], data_dict['X_col']
    assert len(X_row) == X.shape[0] and len(X_col) == X.shape[1]

    y = data_dict['y'].astype('float32')
    y_row, y_col = data_dict['y_row'], data_dict['y_col']
    assert len(y_row) == y.shape[0] and len(y_col) == y.shape[1]

    assert np.all(X_row == y_row), \
    "Cell order mismatch between gene count and protein count"

    # ====== filter zero columns ====== #
    X, X_col = remove_allzeros_columns(matrix=X,
                                       colname=X_col,
                                       print_log=verbose)

    save_to_dataset(path=preprocessed_path,
                    X=X,
                    X_col=X_col,
                    y=y,
                    y_col=y_col,
                    rowname=X_row,
                    print_log=verbose)
  # ******************** read preprocessed data ******************** #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds


# ===========================================================================
# Full FACS data
# ===========================================================================
def read_full_FACS(override=False, verbose=False):
  """ https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE75478
  This is the full FACS data of 2 individuals with 7 protein markers
  """
  download_path = os.path.join(DOWNLOAD_DIR, "FACS_full")
  if not os.path.exists(download_path):
    os.mkdir(download_path)
  # ====== download the data ====== #
  file_url = [
      ('GSE75478_transcriptomics_facs_indeces_filtered_I1.csv.gz',
       'https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE75478&format=file&file=GSE75478%5Ftranscriptomics%5Ffacs%5Findeces%5Ffiltered%5FI1%2Ecsv%2Egz'
      ),
      ('GSE75478_transcriptomics_facs_indeces_filtered_I2.csv.gz',
       'https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE75478&format=file&file=GSE75478%5Ftranscriptomics%5Ffacs%5Findeces%5Ffiltered%5FI2%2Ecsv%2Egz'
      ),
      ('GSE75478_transcriptomics_raw_filtered_I1.csv.gz',
       'https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE75478&format=file&file=GSE75478%5Ftranscriptomics%5Fraw%5Ffiltered%5FI1%2Ecsv%2Egz'
      ),
      ('GSE75478_transcriptomics_raw_filtered_I2.csv.gz',
       'https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE75478&format=file&file=GSE75478%5Ftranscriptomics%5Fraw%5Ffiltered%5FI2%2Ecsv%2Egz'
      ),
  ]
  for name, url in file_url:
    if not os.path.exists(os.path.join(download_path, name)):
      get_file(fname=name, origin=url, outdir=download_path, verbose=verbose)
  # ====== extract the data ====== #
  preprocessed_path = _FACS_PREPROCESSED % 7
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)
  elif override:
    shutil.rmtree(preprocessed_path)
    os.mkdir(preprocessed_path)
  # ******************** preprocessed data NOT found ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    data_map = {}
    for name, _ in file_url:
      zip_path = os.path.join(download_path, name)
      with gzip.open(zip_path, 'rb') as f:
        data_map[name.split('.')[0]] = np.array(
            [str(line, 'utf-8').strip().split(',') for line in f]).T

    i1 = data_map['GSE75478_transcriptomics_raw_filtered_I1']
    f1 = data_map['GSE75478_transcriptomics_facs_indeces_filtered_I1']

    i2 = data_map['GSE75478_transcriptomics_raw_filtered_I2']
    f2 = data_map['GSE75478_transcriptomics_facs_indeces_filtered_I2']
    # Matching duplicated row in `i` and `f`
    row_name = set(i1[1:, 0]) & set(f1[1:, 0])
    i1 = i1[[True] + [True if i in row_name else False for i in i1[1:, 0]], :]
    f1 = f1[[True] + [True if i in row_name else False for i in f1[1:, 0]], :]
    assert np.all(i1[:, 0] == f1[:, 0])

    row_name = set(i2[1:, 0]) & set(f2[1:, 0])
    i2 = i2[[True] + [True if i in row_name else False for i in i2[1:, 0]], :]
    f2 = f2[[True] + [True if i in row_name else False for i in f2[1:, 0]], :]
    assert np.all(i2[:, 0] == f2[:, 0])

    # Matching the genes and protein among individuals
    gene_name = set(i1[0][1:]) & set(i2[0][1:])
    i1 = i1[:, [True] + [True if i in gene_name else False for i in i1[0][1:]]]
    i2 = i2[:, [True] + [True if i in gene_name else False for i in i2[0][1:]]]
    assert np.all(i1[0] == i2[0])
    gene = np.concatenate((i1, i2[1:]), axis=0)

    prot_name = set([i for i in set(f1[0][1:]) & set(f2[0][1:]) if '_cd' in i])
    prot_name = sorted(prot_name)
    f1 = f1[:, [0] + [f1[0].tolist().index(i) for i in prot_name]]
    f2 = f2[:, [0] + [f2[0].tolist().index(i) for i in prot_name]]
    assert np.all(f1[0] == f2[0])
    prot = np.concatenate((f1, f2[1:]), axis=0)

    # ====== save data to disk ====== #
    X = gene[1:, 1:].astype('float32')
    X_row = gene[1:, 0]
    X_col = gene[0, 1:]
    X_col = np.array([i.replace('"', '') for i in X_col])

    y = prot[1:, 1:].astype('float32')
    y_row = prot[1:, 0]
    y_col = np.array(
        [i.replace('"', '').split('_')[-1].upper() for i in prot[0, 1:]])

    assert np.all(X_row == y_row)
    X_row = np.array([i.replace('"', '') for i in X_row])

    # ====== the protein marker can be smaller than zero ====== #
    min_values = np.min(y, axis=0, keepdims=True)
    min_values = np.where(min_values > 0, 0, min_values)
    y = y + np.abs(min_values)
    # ====== filter zero columns ====== #
    X, X_col = remove_allzeros_columns(matrix=X,
                                       colname=X_col,
                                       print_log=verbose)
    save_to_dataset(path=preprocessed_path,
                    X=X,
                    X_col=X_col,
                    y=y,
                    y_col=y_col,
                    rowname=X_row,
                    print_log=verbose)
  # ******************** read preprocessed data ******************** #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds



# add path in order of priority
# DEFAULT_BASE_DIR: default prefix path if no configuration is found
# DATA_DIR/PREPROCESSED_BASE_DIR: path for storing preprocessed data
# DOWNLOAD_DIR: path for storing downloaded (original) data
# EXP_DIR: path for storing experiments results
import os
from os.path import expanduser

from odin.utils import select_path

DEFAULT_BASE_DIR = expanduser("~")

if 'SISUA_DATA' in os.environ:
  DATA_DIR = os.environ['SISUA_DATA']
  if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)
  elif os.path.isfile(DATA_DIR):
    raise RuntimeError("Store data path at '%s' must be a folder" % DATA_DIR)
else:
  DATA_DIR = select_path(os.path.join(DEFAULT_BASE_DIR, 'bio_data'),
                         create_new=True)

PREPROCESSED_BASE_DIR = DATA_DIR
DOWNLOAD_DIR = select_path(os.path.join(DATA_DIR, 'downloads'), create_new=True)

# PATH for saving experiments results
if 'SISUA_EXP' in os.environ:
  EXP_DIR = os.environ['SISUA_EXP']
  if not os.path.exists(EXP_DIR):
    os.mkdir(EXP_DIR)
  elif os.path.isfile(EXP_DIR):
    raise RuntimeError("Experiment path at '%s' must be a folder" % EXP_DIR)
else:
  EXP_DIR = select_path(os.path.join(DEFAULT_BASE_DIR, 'bio_log'),
                        create_new=True)



from __future__ import print_function, division, absolute_import

import os
import time
import pickle
from six import string_types
from multiprocessing import Process
from itertools import product

import seaborn as sns
from matplotlib import pyplot as plt

from odin.utils import as_tuple, ctext, catch_warnings_ignore, ArgController

import numpy as np

# ===========================================================================
# Argument helper
# ===========================================================================
def get_arguments():
  args = ArgController(
  ).add('-ds',
        "name of multiple datasets for cross-data analysis: 'cross8k_ly,crossecc_ly'",
        "cross8k_ly,crossecc_ly"
  ).add("-model",
        "name of model for testing, for example: 'vae', 'movae', 'vae,movae'",
        "movae"
  ).add("-path",
        "Saving all figures to given output folder",
        "/tmp/cross_analysis"
  ).add("--verbose",
        "Enable verbose logging",
        False
  ).add("-nprocess",
        "Number of multiple processes for running the experiments",
        2
  ).parse()
  datasets = [i.strip().lower()
              for i in str(args.ds).split(',')
              if len(i.strip()) > 0]
  model = [i.strip().lower()
           for i in str(args.model).split(',')
           if len(i.strip()) > 0]
  nprocess = int(args.nprocess)
  return dict(datasets=datasets, models=model,
              outpath=args.path, verbose=bool(args.verbose),
              nprocess=nprocess)


# ===========================================================================
# Main methods
# ===========================================================================
all_datasets = None

def _analyze(ds_name, model_path, outpath,
             y_true, all_proteins, verbose):
  from sisua.analysis import Posterior
  with open(model_path, 'rb') as f:
    infer = pickle.load(f)
  ds_infer = infer.configs['dataset']
  ds = [j for i, j in all_datasets if i == ds_name][0]

  # path is a folder
  path = os.path.join(outpath, 'data%s_model%s' %
    (ds_name.replace('_', '').upper(),
     ds_infer.replace('_', '').upper()))
  path = os.path.join(path, infer.short_id)
  if not os.path.exists(path):
    os.mkdir(path)

  # log start
  if verbose:
    print("\nData:%s - Model:%s" % (
        ctext(ds_name, 'yellow'),
        ctext(ds_infer, 'yellow')))
    print(" Outpath:", ctext(path, 'cyan'))

  # create a mixed Posterior
  pos = Posterior(infer, ds=ds)
  # a lot of figures so RuntimeWarning about maximum amount
  # of figure will be appeared
  with catch_warnings_ignore(RuntimeWarning):
    # analysis
    pos.new_figure(
    ).plot_latents_binary_scatter(size=4
    ).plot_latents_distance_heatmap(
    ).plot_correlation_marker_pairs()
    # protein series
    if infer.is_semi_supervised:
      y_pred = {i: j
                for i, j in zip(dict(all_datasets)[ds_infer]['y_col'],
                                infer.predict_y(ds['X']).T)
                if i in all_proteins}
      y_pred = np.hstack([y_pred[i][:, np.newaxis]
                          for i in all_proteins])
      pos.plot_protein_predicted_series(
          y_true_new=y_true, y_pred_new=y_pred, labels_new=all_proteins)
      for prot_name in all_proteins:
        pos.plot_protein_scatter(protein_name=prot_name,
          y_true_new=y_true, y_pred_new=y_pred, labels_new=all_proteins)
    # save plot and show log
    pos.save_plots(path, dpi=80)


def cross_analyze(datasets, outpath, models,
                  nprocess=1, verbose=False):
  from sisua.data import get_dataset
  from sisua.data.path import EXP_DIR
  from sisua.data.utils import standardize_protein_name

  assert nprocess > 0, "Number of processes must be greater than 0"

  datasets = as_tuple(datasets, t=string_types)
  assert len(datasets) > 1, \
  "Require more than one datasets for cross analysis"
  if not os.path.exists(outpath):
    os.mkdir(outpath)

  models = as_tuple(models, t=string_types)
  assert len(models) > 0, \
  "At least one model must be given"

  # ====== load datasets ====== #
  global all_datasets
  all_datasets = {name: get_dataset(name)[0]
                  for name in datasets}
  all_datasets = [
  (name, dict(X=ds['X'][:],
              X_col=ds['X_col'],
              X_row=ds['X_row'],
              y=ds['y'],
              y_col=np.array([standardize_protein_name(i)
                              for i in ds['y_col']]),
  ))
  for name, ds in all_datasets.items()]

  # ====== check gene expression is matching ====== #
  genes = all_datasets[0][1]['X_col']
  for name, ds in all_datasets:
    assert np.all(ds['X_col'] == genes), "Set of training genes mis-match"

  # ====== get the list of all overlapping protein ====== #
  all_proteins = set(all_datasets[0][1]['y_col'])
  for name, ds in all_datasets:
    all_proteins &= set(ds['y_col'])
  all_proteins = sorted(all_proteins)

  # ====== only select certain protein ====== #
  if verbose:
    print("Datasets       :", ctext(', '.join(datasets), 'yellow'))
    print("Model          :", ctext(', '.join(models), 'yellow'))
    print("Shared proteins:", ctext(', '.join(all_proteins), 'yellow'))
    for name, ds in all_datasets:
      print(" ", ctext(name, 'cyan'))
      print("   X    :", ds['X'].shape)
      print("   X_col:", ds['X_col'])
      print("   y    :", ds['y'].shape)
      print("   y_col:", ', '.join(ds['y_col']))

  # ====== load all the model ====== #
  all_models = []
  for ds_name in datasets:
    if verbose:
      print("Search model for dataset '%s' ..." % ctext(ds_name, 'yellow'))
    exp_path = os.path.join(EXP_DIR, ds_name)
    for model_name in os.listdir(exp_path):
      if model_name.split('_')[0] in models:
        path = os.path.join(exp_path, model_name, 'model.pkl')
        if os.path.exists(path):
          all_models.append(path)
          if verbose:
            print(" ", ctext(model_name, 'cyan'))
  if verbose:
    print("%s datasets and %s models => %s experiments" % (
        ctext(len(all_datasets), 'yellow'),
        ctext(len(all_models), 'yellow'),
        ctext(len(all_datasets) * len(all_models), 'yellow'),
    ))

  # ====== create all necessary dir in advance ====== #
  all_data_name = [i[0] for i in all_datasets]
  all_model_name = [i.split('/')[-3] for i in all_models]
  for name1, name2 in product(all_data_name, all_model_name):
    path = os.path.join(outpath, 'data%s_model%s' %
      (name1.replace('_', '').upper(),
       name2.replace('_', '').upper()))
    if not os.path.exists(path):
      os.mkdir(path)
      if verbose:
        print("Create output folder:", ctext(path, 'yellow'))

  # ====== start generate analysis ====== #
  processes = []
  for ds_name, ds in all_datasets:
    y_true = {i: j
              for i, j in zip(ds['y_col'], ds['y'].T)
              if i in all_proteins}
    # preserve the same order of all_proteins
    y_true = np.hstack([y_true[i][:, np.newaxis]
                        for i in all_proteins])

    for model_path in all_models:
      processes.append(
          Process(
              target=_analyze,
              args=(ds_name, model_path, outpath, y_true, all_proteins, verbose))
      )
      if len(processes) >= nprocess:
        [p.start() for p in processes]
        [p.join() for p in processes]
        processes = []

  # finish the remain processes
  if len(processes) > 0:
    [p.start() for p in processes]
    [p.join() for p in processes]

# ===========================================================================
# Main
# ===========================================================================
if __name__ == '__main__':
  cross_analyze(**get_arguments())



# Reference:
# "Single-cell transcriptomics reveals expansion of cytotoxic CD4 T-cells in
# supercentenarians", Kosuke Hashimoto, Tsukasa Kouno, Tomokatsu Ikawa, el. at
# https://www.biorxiv.org/content/10.1101/643528v1
import gzip
import os
import pickle
import shutil

import numpy as np

from odin.fuel import Dataset, MmapArrayWriter
from odin.stats import describe
from odin.utils import batching, ctext, get_file, one_hot, select_path
from odin.utils.crypto import decrypt_aes, md5_checksum
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR
from sisua.data.utils import remove_allzeros_columns, save_to_dataset

path = "/home/trung/bio_data/downloads/SuperCentenarian_original/01.UMI.txt.gz"

# ===========================================================================
# Constants
# ===========================================================================
# ## ---- Raw UMI counts ---- ##
# 01.UMI.txt.gz
#         row: Ensembl Gene ID (23384 genes)
#         col: Cell barcode (61202 cells)
# ## ---- Normalized expression ---- ##
# 02.UMI.lognorm.txt.gz
#         row: Ensembl Gene ID (23384 genes)
#         col: Cell barcode (61202 cells)
# ## ---- Table of Sample - Cell barcode ---- ##
# 03.Cell.Barcodes.txt.gz
#         Col1: Cell barcode (61202 cells)
#         Col2: Sample ID
#         Col3: Sample Type (SC and CT denotes supercentenarians and controls)
# ## ---- The second sequencing of SC1 ---- ##
# 04.SC1.tar
#         barcodes.tsv.gz: Cell barcode
#         features.tsv.gz: Ensemble Gene ID
#         matrix.mtx.gz: pression matrix in the Market Exchange Format
# ## ---- The second sequencing of SC2 ---- ##
# 05.SC2.tar
#         barcodes.tsv.gz: Cell barcode
#         features.tsv.gz: Ensemble Gene ID
#         matrix.mtx.gz: pression matrix in the Market Exchange Format
# Protein
_URL = [
    "http://gerg.gsc.riken.jp/SC2018/01.UMI.txt.gz",
    "http://gerg.gsc.riken.jp/SC2018/02.UMI.lognorm.txt.gz",
    "http://gerg.gsc.riken.jp/SC2018/03.Cell.Barcodes.txt.gz",
    "http://gerg.gsc.riken.jp/SC2018/04.SC1.tar",
    "http://gerg.gsc.riken.jp/SC2018/05.SC2.tar",
]


def read_gzip_csv(path):
  data = []
  cell_id = None
  gene_id = None
  with gzip.open(path, mode='rb') as f:
    for line in f:
      line = str(line, 'utf-8').strip().split('\t')
      data.append(line)
  cell_id = np.array(data[0])
  data = data[1:]
  gene_id = np.array([i[0] for i in data])
  data = np.array([i[1:] for i in data], dtype='float32').T
  return data, cell_id, gene_id


# ===========================================================================
# Main
# ===========================================================================
def read_centenarian(override=False, verbose=False):
  download_path = os.path.join(DOWNLOAD_DIR, "SuperCentenarian_original")
  if not os.path.exists(download_path):
    os.mkdir(download_path)

  preprocessed_path = os.path.join(PREPROCESSED_BASE_DIR,
                                   'SuperCentenarian_preprocessed')

  if override and os.path.exists(preprocessed_path):
    shutil.rmtree(preprocessed_path)
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)

  # ******************** preprocessed ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    labels = get_file(fname=os.path.basename(_URL[2]),
                      origin=_URL[2],
                      outdir=download_path,
                      verbose=verbose)

    data = []
    with gzip.open(labels, mode='rb') as f:
      for line in f:
        line = str(line, 'utf-8').strip().split('\t')
        assert line[1][:2] == line[2]
        data.append(line)
    labels = np.array(data)
    y_col = sorted(set(labels[:, 1]))
    y = one_hot(np.array([y_col.index(i) for i in labels[:, 1]]),
                len(y_col)).astype('float32')
    y_col = np.array(y_col)

    raw = get_file(fname=os.path.basename(_URL[0]),
                   origin=_URL[0],
                   outdir=download_path,
                   verbose=verbose)
    if verbose:
      print("Unzip and reading raw UMI ...")
    X_raw, cell_id1, gene_id1 = read_gzip_csv(raw)

    norm = get_file(fname=os.path.basename(_URL[1]),
                    origin=_URL[1],
                    outdir=download_path,
                    verbose=verbose)
    if verbose:
      print("Unzip and reading log-norm UMI ...")
    X_norm, cell_id2, gene_id2 = read_gzip_csv(norm)

    assert np.all(cell_id1 == cell_id2) and np.all(labels[:, 0] == cell_id1) and \
      np.all(gene_id1 == gene_id2)
    assert X_raw.shape[0] == X_norm.shape[0] == len(cell_id1) and \
      X_raw.shape[1] == X_norm.shape[1] == len(gene_id1)

    if verbose:
      print("Saving data to %s ..." % ctext(preprocessed_path, 'cyan'))
    save_to_dataset(preprocessed_path,
                    X=X_raw,
                    X_col=gene_id1,
                    y=y,
                    y_col=y_col,
                    rowname=cell_id1,
                    print_log=verbose)
    with MmapArrayWriter(os.path.join(preprocessed_path, 'X_log'),
                         shape=(0, X_norm.shape[1]),
                         dtype='float32',
                         remove_exist=True) as f:
      for s, e in batching(batch_size=2048, n=X_norm.shape[0]):
        f.write(X_norm[s:e])
  # ====== read preprocessed data ====== #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds



# Tensorflow re-implementaiton of: https://github.com/YosefLab/scVI/tree/master/scvi
# MIT License
# Copyright (c) 2018, Romain Lopez
# Modification by Trung Ngo 2019
from __future__ import absolute_import, division, print_function

import numpy as np
import tensorflow as tf
from tensorflow.python.keras.layers import Dense
from tensorflow_probability.python.distributions import Normal

from odin.bay.distribution_layers import (NegativeBinomialDispLayer,
                                          ZINegativeBinomialDispLayer)
from odin.bay.helpers import Statistic
from odin.networks import DenseDistribution, Identity
from sisua.models.base import SingleCellModel
from sisua.models.latents import get_latent
from sisua.models.networks import DenseNetwork


class SCVI(SingleCellModel):
  """ Re-implementation of single cell variational inference in tensorflow

  Parameters
  ----------
  dispersion : {'full', 'share', 'single'}
    'full' is 'gene-cell' mode in scVI - dispersion can differe for every gene
    in every cell
    'share' is 'gene' mode in scVI - dispersion is constant per gene across
    cells
    'single' - single value dispersion for all genes and cells (only in SISUA)
  clip_library : `float` (default=`10`)
    clipping the maximum library size to prevent overflow in exponential,
    e.g. if L=10 then the maximum library value is exp(10)=~22000
  """

  def __init__(self,
               units,
               dispersion='full',
               xdist='zinbd',
               zdist='normal',
               ldist='normal',
               xdrop=0.3,
               edrop=0,
               zdrop=0,
               ddrop=0,
               hdim=128,
               zdim=32,
               nlayers=2,
               clip_library=12,
               batchnorm=True,
               linear_decoder=False,
               **kwargs):
    super(SCVI, self).__init__(units=units,
                               xdist=xdist,
                               dispersion=dispersion,
                               parameters=locals(),
                               **kwargs)
    assert self.xdist[0] in (NegativeBinomialDispLayer,
                             ZINegativeBinomialDispLayer), \
      "scVI only suppoert 'nbd' and 'zinbd' as output distribution, but given:"+\
        "%s" % str(self.xdist[0])
    # initialize the autoencoder
    self.encoder_z = DenseNetwork(n_units=hdim,
                                  nlayers=nlayers,
                                  activation='relu',
                                  batchnorm=batchnorm,
                                  input_dropout=xdrop,
                                  output_dropout=edrop,
                                  seed=self.seed,
                                  name='EncoderZ')
    self.encoder_l = DenseNetwork(n_units=1,
                                  nlayers=1,
                                  activation='relu',
                                  batchnorm=batchnorm,
                                  input_dropout=xdrop,
                                  output_dropout=edrop,
                                  seed=self.seed,
                                  name='EncoderL')

    if linear_decoder:
      self.decoder = Identity(name='LinearDecoder')
    else:
      self.decoder = DenseNetwork(n_units=hdim,
                                  nlayers=nlayers,
                                  activation='relu',
                                  batchnorm=batchnorm,
                                  input_dropout=zdrop,
                                  output_dropout=ddrop,
                                  seed=self.seed,
                                  name='Decoder')
    self.latent = get_latent(zdist)(units=zdim, name=self.name + 'Latent')
    self.library = get_latent(ldist)(units=1, name=self.name + 'Library')
    self.clip_library = float(clip_library)
    n_dims = self.units[0]
    # mean gamma (logits value, applying softmax later)
    self.px_scale = Dense(units=n_dims, activation='linear', name="MeanScale")
    # dropout logits value
    if self.is_zero_inflated:
      self.px_dropout = Dense(n_dims, activation='linear', name="DropoutLogits")
    else:
      self.px_dropout = Identity(name="DropoutLogits")
    # dispersion
    if self.dispersion == 'full':
      self.px_r = Dense(n_dims, activation='linear', name='Dispersion')
    elif self.dispersion == 'share':
      self.px_r = self.add_weight(name='Dispersion',
                                  shape=(1, n_dims),
                                  initializer=tf.initializers.RandomNormal,
                                  trainable=True)
    else:
      self.px_r = self.add_weight(name='Dispersion',
                                  shape=(),
                                  initializer=tf.initializers.RandomNormal,
                                  trainable=True)
    # mRNA distribution
    self.pX = self.xdist[0](n_dims, given_log_mean=False, given_log_disp=False)
    # semi-supervised distributions
    for idx, (units, posterior, activation) in enumerate(
        zip(self.units[1:], self.xdist[1:], self.xactiv[1:])):
      post = DenseDistribution(
          units=units,
          posterior=posterior,
          activation=activation,
          posterior_kwargs=dict(dispersion=self.dispersion))
      setattr(self, 'output_layer%d' % idx, post)

  def _call(self, x, lmean, lvar, t, y, masks, training, n_samples):
    # applying encoding
    e_z = self.encoder_z(x, training=training)
    e_l = self.encoder_l(x, training=training)

    # latent spaces
    qZ = self.latent(e_z, mode=Statistic.DIST)
    qL = self.library(e_l, mode=Statistic.DIST)
    Z_samples = qZ.sample(n_samples)
    # clipping L value to avoid overflow, exp(12) = ~160000
    L_samples = qL.sample(n_samples)
    L_samples = tf.clip_by_value(L_samples, 0, self.clip_library)
    # decoding the latent
    d = self.decoder(Z_samples, training=training)

    # mean parameterizations
    px_scale = tf.nn.softmax(self.px_scale(d), axis=1)
    px_scale = tf.clip_by_value(px_scale, 0 + 1e-8, 1 - 1e-8)
    px_rate = tf.math.exp(L_samples) * px_scale
    # dispersion parameterizations
    if self.dispersion == 'full':
      px_r = self.px_r(d)
    else:
      px_r = self.px_r
    px_r = tf.math.exp(px_r)
    # dropout for zero inflation
    px_dropout = self.px_dropout(d)

    # mRNA expression distribution
    # this order is the same as how the parameters are splited in
    # distribution layer
    if self.is_zero_inflated:
      pX = self.pX(tf.concat((px_rate, px_r, px_dropout), axis=-1))
    else:
      pX = self.pX(tf.concat((px_rate, px_r), axis=-1))

    # for semi-supervised learning
    pY = []
    for idx in range(self.n_outputs - 1):
      name = 'output_layer%d' % idx
      post = getattr(self, name)
      pY.append(post(d, mode=Statistic.DIST))

    # log like-lihood
    llk_x = tf.expand_dims(pX.log_prob(t), -1)
    # semi-supervised NLLK
    loss_y = tf.convert_to_tensor(0, dtype=x.dtype)
    for i_true, m, i_pred, fn_loss in zip(y, masks, pY, self.xloss[1:]):
      loss_y += fn_loss(i_true, i_pred) * m

    # KL for both latent and library
    kl_z = self.latent.kl_divergence(analytic_kl=self.kl_analytic,
                                     n_samples=n_samples)
    kl_l = self.library.kl_divergence(analytic_kl=self.kl_analytic,
                                      n_samples=n_samples,
                                      prior=Normal(loc=lmean,
                                                   scale=tf.math.sqrt(lvar)))
    # for semi-superivsed, we have the negative log-likelihood,
    # hence the minus sign
    elbo = llk_x - loss_y - kl_l - kl_z * self.kl_weight
    elbo = tf.reduce_logsumexp(elbo, axis=0)
    loss = tf.reduce_mean(-elbo)
    if training:
      self.add_loss(loss)

    # NOTE: add_metric should not be in control if-then-else
    self.add_metric(tf.reduce_max(L_samples), aggregation='mean', name="Lmax")
    self.add_metric(tf.reduce_mean(kl_z), aggregation='mean', name="KLqpZ")
    self.add_metric(tf.reduce_mean(kl_l), aggregation='mean', name="KLqpL")
    self.add_metric(tf.reduce_mean(-llk_x), aggregation='mean', name="nllk_x")
    if self.is_semi_supervised:
      self.add_metric(tf.reduce_mean(loss_y), aggregation='mean', name="nllk_y")
    return [pX] + pY, (qZ, qL)



from __future__ import absolute_import, division, print_function

from abc import ABCMeta, abstractmethod
from enum import Flag, auto

import numpy as np
import scanpy as sc
from six import add_metaclass

from sisua.data.single_cell_dataset import SingleCellOMIC


# ===========================================================================
# Helpers
# ===========================================================================
class Methods(Flag):
  RAW = auto()
  CELL_COUNT = auto()
  GENE_COUNT = auto()
  TOTAL_COUNTS = auto()
  GENE_DISP = auto()
  EXP = auto()
  LOG = auto()
  SCALE = auto()
  PROB = auto()
  BIN = auto()

  @property
  def is_filtering(self):
    """ Return `True` if the normalization method contain any filter for
    cells or genes """
    if any(i in self for i in (Methods.CELL_COUNT, Methods.GENE_COUNT,
                               Methods.GENE_DISP)):
      return True
    return False

  @property
  def is_probabilistic_embedding(self):
    return Methods.PROB in self or Methods.BIN in self

  def __len__(self):
    count = 0
    for i in list(Methods)[1:]:
      if i in self:
        count += 1
    return count

  def __call__(self, X, **kwargs):
    return self.apply(X, **kwargs)

  def __iter__(self):
    for method in list(Methods)[1:]:
      if method in self:
        yield method

  def apply(self,
            X,
            min_counts_per_gene=1,
            min_cells_per_gene=None,
            min_counts_per_cell=1,
            min_genes_per_cell=None,
            min_disp=0.5,
            flavor='seurat',
            n_top_genes=None,
            target_sum=None,
            max_value=None,
            gene_subset=None,
            trained_embedding=None):
    assert isinstance(X, SingleCellOMIC), \
      "X must be instance of sisua.data.SingleCellOMIC"
    is_logarithmized = False
    for method in self:
      # cell filtering
      if method == Methods.CELL_COUNT:
        X.filter_cells(min_counts=min_counts_per_cell,
                       min_genes=min_genes_per_cell,
                       inplace=True)

      # gene filtering (cell countsums normalization is included)
      if gene_subset is None:
        if method == Methods.GENE_COUNT:
          X.filter_genes(min_counts=min_counts_per_gene,
                         min_cells=min_cells_per_gene,
                         inplace=True)
        if method == Methods.TOTAL_COUNTS:
          X.normalize(total_counts=True, target_sum=target_sum, inplace=True)
        if method == Methods.GENE_DISP:
          # Expects logarithmized data.
          X.normalize(log1p=True, inplace=True)
          is_logarithmized = True
          X.filter_highly_variable_genes(min_disp=min_disp,
                                         n_top_genes=n_top_genes,
                                         flavor=flavor,
                                         inplace=True)
      else:
        if method == Methods.TOTAL_COUNTS:
          X.normalize(total_counts=True, target_sum=target_sum, inplace=True)
        if X.shape[1] > len(gene_subset):
          gene_ids = gene_subset.index.astype('int')
          X.apply_indices(gene_ids, observation=False)

      # highly variable genes
      if method == Methods.EXP:
        X.expm1(inplace=True)
      elif method == Methods.LOG:
        if not is_logarithmized:
          X.normalize(log1p=True, inplace=True)
      elif method == Methods.SCALE:
        X.normalize(scale=True, max_value=max_value, inplace=True)
      elif method == Methods.PROB or method == Methods.BIN:
        X.probabilistic_embedding(pbe=trained_embedding)
        X.obsm["X_org"] = X._X
        if method == Methods.PROB:
          X._X = X.obsm['X_prob']
          X._name = X._name[:-3] + 'prob'
        else:
          X._X = X.obsm['X_bin']
          X._name = X._name[:-3] + 'bin'
    return X


# ===========================================================================
# Recipes
# ===========================================================================
class NormalizationRecipe(object):
  """ Predefined normalization recipe

  Note
  ----
  It is recommended to enable corruption within this normalization to
  simulate actual conditions
  """

  def __init__(self,
               methods: Methods = Methods.RAW,
               corruption_rate=0.25,
               corruption_dist='binomial',
               target_sum=None,
               max_value=None,
               min_counts_per_gene=1,
               min_cells_per_gene=None,
               min_counts_per_cell=1,
               min_genes_per_cell=None,
               min_disp=0.5,
               flavor='seurat',
               n_top_genes=None):
    super(NormalizationRecipe, self).__init__()
    # pylint just could not detect self.__dict__.update(locals())
    # so we assign each argument one-by-one
    self.methods = methods
    self.corruption_rate = corruption_rate
    self.corruption_dist = corruption_dist
    self.target_sum = target_sum
    self.max_value = max_value
    self.min_counts_per_gene = min_counts_per_gene
    self.min_cells_per_gene = min_cells_per_gene
    self.min_counts_per_cell = min_counts_per_cell
    self.min_genes_per_cell = min_genes_per_cell
    self.min_disp = min_disp
    self.flavor = flavor
    self.n_top_genes = n_top_genes
    # validate some important type
    assert isinstance(self.methods, Methods), \
      'methods must be instance of sisua.data.normalization_recipes.Methods'
    # some statistics might be stored during training
    self._is_trained = False
    self._gene_subset = None
    self._trained_embedding = None

  @property
  def name(self):
    return self.__class__.__name__.lower()

  @property
  def is_trained(self):
    return self._is_trained

  def __call__(self, X, training=False, copy=True):
    return self.normalize(X, training=training, copy=copy)

  def normalize(self, X, training=False, copy=True):
    """
    X : SingleCellOMIC
    training : data is used for training or not (
      corruption is enable for training)
    copy : create a copy version of data for normalization
    """
    if isinstance(X, SingleCellOMIC):
      X = X.copy() if copy else X
    elif isinstance(X, sc.AnnData):
      X = SingleCellOMIC(X.copy() if copy else X)
    else:
      X = SingleCellOMIC(X=X)

    # data should be corrupted before doing any preprocessing to
    # match the actual condition
    if training and self.corruption_rate > 0:
      X.corrupt(corruption_rate=self.corruption_rate,
                corruption_dist=self.corruption_dist,
                inplace=True)

    X = self.methods.apply(X,
                           min_counts_per_gene=self.min_counts_per_gene,
                           min_cells_per_gene=self.min_cells_per_gene,
                           min_counts_per_cell=self.min_counts_per_cell,
                           min_genes_per_cell=self.min_genes_per_cell,
                           min_disp=self.min_disp,
                           flavor=self.flavor,
                           n_top_genes=self.n_top_genes,
                           target_sum=self.target_sum,
                           max_value=self.max_value,
                           gene_subset=self._gene_subset,
                           trained_embedding=self._trained_embedding)

    if training and not self.is_trained:
      self._is_trained = True
      # assum the first var column is always geneID
      self._gene_subset = X.var.iloc[:, 0]
      if 'pbe' in X.uns:
        self._trained_embedding = X.uns['pbe']
    return X

  def __str__(self):
    return self.name


class Seurat(NormalizationRecipe):
  """Normalization and filtering as of Seurat [Satija15]_.
  This uses a particular preprocessing.
  Expects non-logarithmized data.

  Note
  ----
  It is recommended to enable corruption within this normalization to
  simulate actual conditions
  """

  def __init__(self,
               corruption_rate=0.25,
               corruption_dist='binomial',
               min_cells=3,
               min_genes=200,
               log=True):
    super(Seurat,
          self).__init__(methods=Methods.CELL_COUNT | Methods.GENE_COUNT |
                         Methods.TOTAL_COUNTS | Methods.GENE_DISP |
                         (Methods.LOG if log else Methods.EXP) | Methods.SCALE,
                         target_sum=1e4,
                         max_value=10,
                         min_counts_per_gene=None,
                         min_cells_per_gene=min_cells,
                         min_counts_per_cell=None,
                         min_genes_per_cell=min_genes,
                         min_disp=0.5,
                         flavor='seurat',
                         n_top_genes=None,
                         corruption_rate=corruption_rate,
                         corruption_dist=corruption_dist)


class CellRanger(NormalizationRecipe):
  """Normalization and filtering as of [Zheng17]_.
  Reproduces the preprocessing of [Zheng17]_ - the Cell Ranger R Kit of 10x
  Genomics.
  Expects non-logarithmized data.

  The procedure is as follow

  References
  ----------
  pass

  Note
  ----
  It is recommended to enable corruption within this normalization to
  simulate actual conditions
  """

  def __init__(self,
               corruption_rate=0.25,
               corruption_dist='binomial',
               min_counts=1,
               n_top_genes=1000,
               log=True):
    super(CellRanger, self).__init__(methods=Methods.GENE_COUNT |
                                     Methods.TOTAL_COUNTS | Methods.GENE_DISP,
                                     target_sum=None,
                                     max_value=None,
                                     min_counts_per_gene=min_counts,
                                     min_cells_per_gene=None,
                                     min_counts_per_cell=None,
                                     min_genes_per_cell=None,
                                     min_disp=None,
                                     flavor='cell_ranger',
                                     n_top_genes=n_top_genes,
                                     corruption_rate=corruption_rate,
                                     corruption_dist=corruption_dist)
    self.log = bool(log)

  def normalize(self, X, training=False, copy=True):
    """
    X : SingleCellOMIC
    training : data is used for training or not (
      corruption is enable for training)
    copy : create a copy version of data for normalization
    """
    import scanpy as sc
    sc.pp.recipe_seurat
    X = super(CellRanger, self).normalize(X, training=training, copy=copy)
    # renormalize after filtering
    X.expm1()
    X.normalize(total_counts=True, log1p=self.log, scale=True)
    return X


class Sisua(NormalizationRecipe):
  """
  Expects non-logarithmized data.
  """

  def __init__(self,
               corruption_rate=0.25,
               corruption_dist='binomial',
               min_counts=1,
               total_counts=True,
               n_top_genes=1000):
    methods = (Methods.CELL_COUNT | Methods.GENE_COUNT | Methods.TOTAL_COUNTS |
               (Methods.LOG if n_top_genes is None else Methods.GENE_DISP |
                Methods.LOG) | Methods.SCALE)
    if not total_counts:
      methods = methods ^ Methods.TOTAL_COUNTS
    super(Sisua, self).__init__(methods=methods,
                                target_sum=1e4,
                                max_value=None,
                                min_counts_per_gene=min_counts,
                                min_cells_per_gene=None,
                                min_counts_per_cell=min_counts,
                                min_genes_per_cell=None,
                                min_disp=0.5,
                                flavor='cell_ranger',
                                n_top_genes=n_top_genes,
                                corruption_rate=corruption_rate,
                                corruption_dist=corruption_dist)
